/*
 * Copyright (c) 2012 Qualcomm Atheros, Inc.
 * All Rights Reserved.
 * Qualcomm Atheros Confidential and Proprietary.
 * Notifications and licenses are retained for attribution purposes only
 */

/*
 * Copyright (c) 2005, Swedish Institute of Computer Science
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * This file is part of the Contiki operating system.
 *
 * @(#)$Id: //depot/sw/switch/app/elite/trunk/elite/os/contiki/contiki-2.5/core/net/dhcpc.c#16 $
 */

#include <stdio.h>
#include <string.h>

#include "contiki.h"
#include "contiki-net.h"
#include "net/dhcpc.h"
#include "cpu_music.h"

#include "sys_adapt.h"
#include "dhcp_client.h"
#include "elite_sys.h"
#include "elite_ethernet.h"

#define STATE_INITIAL         0
#define STATE_SENDING         1
#define STATE_OFFER_RECEIVED  2
#define STATE_CONFIG_RECEIVED 3

static struct dhcpc_state s;

struct dhcp_msg {
  u8_t op, htype, hlen, hops;
  u8_t xid[4];
  u16_t secs, flags;
  u8_t ciaddr[4];
  u8_t yiaddr[4];
  u8_t siaddr[4];
  u8_t giaddr[4];
  u8_t chaddr[16];
#ifndef UIP_CONF_DHCP_LIGHT
  u8_t sname[64];
  u8_t file[128];
#endif
  u8_t options[312];
};

#define BOOTP_BROADCAST 0x8000

#define DHCP_REQUEST        1
#define DHCP_REPLY          2
#define DHCP_HTYPE_ETHERNET 1
#define DHCP_HLEN_ETHERNET  6
#define DHCP_MSG_LEN      236

#define DHCPC_SERVER_PORT  67
#define DHCPC_CLIENT_PORT  68

#define DHCPDISCOVER  1
#define DHCPOFFER     2
#define DHCPREQUEST   3
#define DHCPDECLINE   4
#define DHCPACK       5
#define DHCPNAK       6
#define DHCPRELEASE   7

#define DHCP_OPTION_SUBNET_MASK   1
#define DHCP_OPTION_ROUTER        3
#define DHCP_OPTION_DNS_SERVER    6
#define DHCP_OPTION_REQ_IPADDR   50
#define DHCP_OPTION_LEASE_TIME   51
#define DHCP_OPTION_MSG_TYPE     53
#define DHCP_OPTION_SERVER_ID    54
#define DHCP_OPTION_REQ_LIST     55
#define DHCP_OPTION_END         255

/* cpu solution */
static UINT8  dhcpc_eth_port = 0;
static UINT16 dhcpc_eth_vlan = 0;

static u32_t xid;
static const u8_t magic_cookie[4] = {99, 130, 83, 99};
/*---------------------------------------------------------------------------*/
static u8_t *
add_msg_type(u8_t *optptr, u8_t type)
{
  *optptr++ = DHCP_OPTION_MSG_TYPE;
  *optptr++ = 1;
  *optptr++ = type;
  return optptr;
}
/*---------------------------------------------------------------------------*/
static u8_t *
add_server_id(u8_t *optptr)
{
  *optptr++ = DHCP_OPTION_SERVER_ID;
  *optptr++ = 4;
  memcpy(optptr, s.serverid, 4);
  return optptr + 4;
}
/*---------------------------------------------------------------------------*/
static u8_t *
add_req_ipaddr(u8_t *optptr)
{
  *optptr++ = DHCP_OPTION_REQ_IPADDR;
  *optptr++ = 4;
  memcpy(optptr, s.ipaddr.u16, 4);
  return optptr + 4;
}
/*---------------------------------------------------------------------------*/
static u8_t *
add_req_options(u8_t *optptr)
{
  *optptr++ = DHCP_OPTION_REQ_LIST;
  *optptr++ = 3;
  *optptr++ = DHCP_OPTION_SUBNET_MASK;
  *optptr++ = DHCP_OPTION_ROUTER;
  *optptr++ = DHCP_OPTION_DNS_SERVER;
  return optptr;
}
/*---------------------------------------------------------------------------*/
static u8_t *
add_end(u8_t *optptr)
{
  *optptr++ = DHCP_OPTION_END;
  return optptr;
}
/*---------------------------------------------------------------------------*/
static void
create_msg(CC_REGISTER_ARG struct dhcp_msg *m)
{
  m->op = DHCP_REQUEST;
  m->htype = DHCP_HTYPE_ETHERNET;
  m->hlen = s.mac_len;
  m->hops = 0;
  memcpy(m->xid, &xid, sizeof(m->xid));
  m->secs = 0;
  m->flags = UIP_HTONS(BOOTP_BROADCAST); /*  Broadcast bit. */
  /*  uip_ipaddr_copy(m->ciaddr, uip_hostaddr);*/
  memcpy(m->ciaddr, uip_hostaddr.u16, sizeof(m->ciaddr));
  memset(m->yiaddr, 0, sizeof(m->yiaddr));
  memset(m->siaddr, 0, sizeof(m->siaddr));
  memset(m->giaddr, 0, sizeof(m->giaddr));
  memcpy(m->chaddr, s.mac_addr, s.mac_len);
  memset(&m->chaddr[s.mac_len], 0, sizeof(m->chaddr) - s.mac_len);
#ifndef UIP_CONF_DHCP_LIGHT
  memset(m->sname, 0, sizeof(m->sname));
  memset(m->file, 0, sizeof(m->file));
#endif

  memcpy(m->options, magic_cookie, sizeof(magic_cookie));
}
/*---------------------------------------------------------------------------*/
static void
send_discover(void)
{
  u8_t *end;
  struct dhcp_msg *m = (struct dhcp_msg *)uip_appdata;

  create_msg(m);

  end = add_msg_type(&m->options[4], DHCPDISCOVER);
  end = add_req_options(end);
  end = add_end(end);

#ifndef WIN32
  elite_ether_forward_type_set(ETH_FWD_TO_ALL_VLAN); /* cpu solution */
#endif

  DHCPC_DEBUG("Send discovery.\r\n");

  uip_send(uip_appdata, (int)(end - (u8_t *)uip_appdata));
}
/*---------------------------------------------------------------------------*/
static void
send_request(void)
{
  u8_t *end;
  struct dhcp_msg *m = (struct dhcp_msg *)uip_appdata;

  create_msg(m);

  end = add_msg_type(&m->options[4], DHCPREQUEST);
  end = add_server_id(end);
  end = add_req_ipaddr(end);
  end = add_end(end);

#ifndef WIN32
  elite_ether_forward_type_set(ETH_FWD_TO_ALL_VLAN); /* cpu solution */
#endif

  DHCPC_DEBUG("Send request.\r\n");

  uip_send(uip_appdata, (int)(end - (u8_t *)uip_appdata));
}
/*---------------------------------------------------------------------------*/
static void
update_conn_remote_ipaddr(uip_ipaddr_t *ripaddr)
{
  memcpy(&(s.conn->ripaddr), ripaddr, sizeof(uip_ipaddr_t));
  //uip_ipaddr_copy(&(s.conn->ripaddr), ripaddr);
}
/*---------------------------------------------------------------------------*/
static void
create_msg_unicast(CC_REGISTER_ARG struct dhcp_msg *m)
{
    m->op = DHCP_REQUEST;
    m->htype = DHCP_HTYPE_ETHERNET;
    m->hlen = s.mac_len;
    m->hops = 0;
    memcpy(m->xid, &xid, sizeof(m->xid));
    m->secs = 0;
    //m->flags = UIP_HTONS(BOOTP_BROADCAST); /*  Broadcast bit. */
    /*  uip_ipaddr_copy(m->ciaddr, uip_hostaddr);*/
    memcpy(m->ciaddr, uip_hostaddr.u16, sizeof(m->ciaddr));
    memset(m->yiaddr, 0, sizeof(m->yiaddr));
    memset(m->siaddr, 0, sizeof(m->siaddr));
    memset(m->giaddr, 0, sizeof(m->giaddr));
    memcpy(m->chaddr, s.mac_addr, s.mac_len);
    memset(&m->chaddr[s.mac_len], 0, sizeof(m->chaddr) - s.mac_len);
#ifndef UIP_CONF_DHCP_LIGHT
    memset(m->sname, 0, sizeof(m->sname));
    memset(m->file, 0, sizeof(m->file));
#endif

    memcpy(m->options, magic_cookie, sizeof(magic_cookie));
}
/*---------------------------------------------------------------------------*/
static void
send_request_renewing(void)
{
  u8_t *end;
  struct dhcp_msg *m = (struct dhcp_msg *)uip_appdata;

  create_msg_unicast(m);

  end = add_msg_type(&m->options[4], DHCPREQUEST);
  end = add_end(end);

#ifndef WIN32
  elite_ether_port_set(dhcpc_eth_port); /* cpu solution */
  elite_ether_vlan_set(dhcpc_eth_vlan); /* cpu solution */
  elite_ether_forward_type_set(ETH_FWD_TO_TARGET_PORT); /* cpu solution */
#endif

  DHCPC_DEBUG("Send request renewing.\r\n");

  uip_send(uip_appdata, (int)(end - (u8_t *)uip_appdata));
}
/*---------------------------------------------------------------------------*/
static u8_t
parse_options(u8_t *optptr, int len)
{
  u8_t *end = optptr + len;
  u8_t type = 0;

  while(optptr < end) {
    switch(*optptr) {
    case DHCP_OPTION_SUBNET_MASK:
      memcpy(s.netmask.u16, optptr + 2, 4);
      break;
    case DHCP_OPTION_ROUTER:
      memcpy(s.default_router.u16, optptr + 2, 4);
      break;
    case DHCP_OPTION_DNS_SERVER:
      memcpy(s.dnsaddr.u16, optptr + 2, 4);
      break;
    case DHCP_OPTION_MSG_TYPE:
      type = *(optptr + 2);
      break;
    case DHCP_OPTION_SERVER_ID:
      memcpy(s.serverid, optptr + 2, 4);
      break;
    case DHCP_OPTION_LEASE_TIME:
      memcpy(s.lease_time, optptr + 2, 4);
      break;
    case DHCP_OPTION_END:
      return type;
    }

    optptr += optptr[1] + 2;
  }
  return type;
}
/*---------------------------------------------------------------------------*/
static int parse_msg(void)
{
    struct dhcp_msg *m = (struct dhcp_msg *)uip_appdata;
    u8_t   allZero[4] = { 0 };

    if (   (m->op == DHCP_REPLY)
        && (memcmp(m->xid, &xid, sizeof(xid)) == 0)
        && (memcmp(m->chaddr, s.mac_addr, s.mac_len) == 0)
        && (memcmp(m->yiaddr, allZero, 4) != 0))
    {
        memcpy(s.ipaddr.u16, m->yiaddr, 4);
        parse_options(&m->options[4], uip_datalen());

        return 0;
    }

    return -1;
}
/*---------------------------------------------------------------------------*/
/*
 * Is this a "fresh" reply for me? If it is, return the type.
 */
static int
msg_for_me(void)
{
  struct dhcp_msg *m = (struct dhcp_msg *)uip_appdata;
  u8_t *optptr = &m->options[4];
  u8_t *end = (u8_t*)uip_appdata + uip_datalen();

  if(m->op == DHCP_REPLY &&
     memcmp(m->xid, &xid, sizeof(xid)) == 0 &&
     memcmp(m->chaddr, s.mac_addr, s.mac_len) == 0) {
    while(optptr < end) {
      if(*optptr == DHCP_OPTION_MSG_TYPE) {
	return *(optptr + 2);
      } else if (*optptr == DHCP_OPTION_END) {
	return -1;
      }
      optptr += optptr[1] + 2;
    }
  }
  return -1;
}

#if ELITE_MODULE_DHCP_BEHAVIOR_1
static u8_t discover_counter = 0;
#endif

/*---------------------------------------------------------------------------*/
unsigned char tx_req_step = 0;
static
PT_THREAD(handle_dhcp(process_event_t ev, void *data))
{
    clock_time_t ticks;

    PT_BEGIN(&s.pt);

init:
    xid++;
    s.state = STATE_SENDING;
#if ELITE_MODULE_DHCP_BEHAVIOR_1
    s.ticks = 4 * CLOCK_SECOND;
    discover_counter = 0;
#else
    s.ticks = CLOCK_SECOND;
#endif

#if ELITE_MODULE_DHCP_BEHAVIOR_1
    while (discover_counter < 15)
#else
    while (1)
#endif
    {
        while (ev != tcpip_event)
        {
            tcpip_poll_udp(s.conn);
            PT_YIELD(&s.pt);
        }

        send_discover();
#if ELITE_MODULE_DHCP_BEHAVIOR_1
        discover_counter++;
#endif
        etimer_set(&s.etimer, s.ticks);

        do
        {
            PT_YIELD(&s.pt);
            if (   (ev == tcpip_event)
                && uip_newdata()
                && (msg_for_me() == DHCPOFFER))
            {
                if (parse_msg() == 0)
                {
                    s.state = STATE_OFFER_RECEIVED;
                    goto selecting;
                }
            }
        } while (!etimer_expired(&s.etimer));

#if (!ELITE_MODULE_DHCP_BEHAVIOR_1)
        if(s.ticks < CLOCK_SECOND * 64)
        {
            s.ticks *= 2;
        }
#endif
    }

#if ELITE_MODULE_DHCP_BEHAVIOR_1
    goto dhcpc_unconfig;
#endif

selecting:
    //xid++;
    s.ticks = CLOCK_SECOND;
    do
    {
        while (ev != tcpip_event)
        {
            tcpip_poll_udp(s.conn);
            PT_YIELD(&s.pt);
        }

        send_request();
        etimer_set(&s.etimer, s.ticks);

        do
        {
            PT_YIELD(&s.pt);
            if (ev == tcpip_event && uip_newdata())
            {
                int msgType = msg_for_me();

                if (msgType == DHCPACK)
                {
                    if (parse_msg() == 0)
                    {
                        s.state = STATE_CONFIG_RECEIVED;

#ifndef WIN32
                        /* cpu solution */
                        dhcpc_eth_port = elite_ether_port_get();
                        dhcpc_eth_vlan = elite_ether_vlan_get();
#endif

                        goto bound;
                    }
                }
                else if (msgType == DHCPNAK)
                {
                    s.state = STATE_INITIAL;
                    goto init;
                }
            }
        } while (!etimer_expired(&s.etimer));

        if (s.ticks < CLOCK_SECOND * 64)
        {
            s.ticks *= 2;
        }
        else
        {
            goto init;
        }
  } while(s.state != STATE_CONFIG_RECEIVED);

bound:
#if 0
    printf("Got IP address %d.%d.%d.%d\n", uip_ipaddr_to_quad(&s.ipaddr));
    printf("Got netmask %d.%d.%d.%d\n",	 uip_ipaddr_to_quad(&s.netmask));
    printf("Got DNS server %d.%d.%d.%d\n", uip_ipaddr_to_quad(&s.dnsaddr));
    printf("Got default router %d.%d.%d.%d\n",
           uip_ipaddr_to_quad(&s.default_router));
    printf("Lease expires in %ld seconds\n",
           uip_ntohs(s.lease_time[0])*65536ul + uip_ntohs(s.lease_time[1]));
#endif

    dhcpc_configured(&s);

#define MAX_TICKS (~((clock_time_t)0) / 2)
#define MAX_TICKS32 (~((u32_t)0))
#define IMIN(a, b) ((a) < (b) ? (a) : (b))

    if (   (uip_ntohs(s.lease_time[0])*65536ul + uip_ntohs(s.lease_time[1]))*CLOCK_SECOND/2
        <= MAX_TICKS32)
    {
        s.ticks =   (uip_ntohs(s.lease_time[0])*65536ul + uip_ntohs(s.lease_time[1]))
                  * CLOCK_SECOND/2;
    }
    else
    {
        s.ticks = MAX_TICKS32;
    }

    while (s.ticks > 0)
    {
        ticks = IMIN(s.ticks, MAX_TICKS);
        s.ticks -= ticks;
        etimer_set(&s.etimer, ticks);
        PT_YIELD_UNTIL(&s.pt, etimer_expired(&s.etimer));
    }

    /* renewing: */
    if(   (uip_ntohs(s.lease_time[0])*65536ul + uip_ntohs(s.lease_time[1]))*CLOCK_SECOND/8*3
       <= MAX_TICKS32)
    {
        s.ticks = (uip_ntohs(s.lease_time[0])*65536ul + uip_ntohs(s.lease_time[1]))*CLOCK_SECOND/8*3;
    }
    else
    {
        s.ticks = MAX_TICKS32;
    }

    xid++;
    {
        uip_ipaddr_t ripaddr;
        uip_ipaddr(&ripaddr, s.serverid[0], s.serverid[1], s.serverid[2], s.serverid[3]);
        update_conn_remote_ipaddr(&ripaddr);
    }
    do
    {
        while (ev != tcpip_event)
        {
            tcpip_poll_udp(s.conn);
            PT_YIELD(&s.pt);
        }
        if(tx_req_step == 0){
            send_request_renewing();
            ticks = 100;
            etimer_set(&s.etimer, ticks);
            tx_req_step = 1;
        }
        else{
            send_request_renewing();
            ticks = IMIN(s.ticks / 2, MAX_TICKS);
            s.ticks -= ticks;
            tx_req_step = 0;
        }
        etimer_set(&s.etimer, ticks);
        do
        {
            PT_YIELD(&s.pt);
            if (ev == tcpip_event && uip_newdata() && msg_for_me() == DHCPACK)
            {
                if (parse_msg() == 0)
                {
                    goto bound;
                }
            }
        } while(!etimer_expired(&s.etimer));
    } while(s.ticks >= CLOCK_SECOND*60);

    // Count down remaining time
    printf("remaining tick: %d (bound: %d)\n", s.ticks, CLOCK_SECOND*60);
    do
    {
        //printf("\r\b count down: %d", s.ticks);
        ticks = 1;
        etimer_set(&s.etimer, ticks);

        do{
           PT_YIELD(&s.pt);
           s.ticks -= 1;
        } while(!etimer_expired(&s.etimer));
    }
    while(s.ticks > 0);

    /* rebinding: */
    tx_req_step = 0;
    if(   (uip_ntohs(s.lease_time[0])*65536ul + uip_ntohs(s.lease_time[1]))*CLOCK_SECOND/8
       <= MAX_TICKS32)
    {
        s.ticks = (uip_ntohs(s.lease_time[0])*65536ul + uip_ntohs(s.lease_time[1]))*CLOCK_SECOND/8;
    }
    else
    {
        s.ticks = MAX_TICKS32;
    }

    xid++;
    {
        uip_ipaddr_t ripaddr;
        uip_ipaddr(&ripaddr, 255, 255, 255, 255);
        update_conn_remote_ipaddr(&ripaddr);
    }
    do
    {
        while (ev != tcpip_event)
        {
            tcpip_poll_udp(s.conn);
            PT_YIELD(&s.pt);
        }
        
        if(tx_req_step == 0){
            send_request_renewing();
            ticks = 100;
            etimer_set(&s.etimer, ticks);
            tx_req_step = 1;
        }
        else{
            send_request_renewing();
            ticks = IMIN(s.ticks / 2, MAX_TICKS);
            s.ticks -= ticks;
            tx_req_step = 0;
        }
        etimer_set(&s.etimer, ticks);
        do
        {
            PT_YIELD(&s.pt);
            if (ev == tcpip_event && uip_newdata() && msg_for_me() == DHCPACK)
            {
                if (parse_msg() == 0)
                {
                    goto bound;
                }
            }
        } while(!etimer_expired(&s.etimer));
    } while(s.ticks >= CLOCK_SECOND*60);

    // Count down remaining time
    do
    {
        ticks = 1;
        etimer_set(&s.etimer, ticks);

        do{
           PT_YIELD(&s.pt);
           s.ticks -= 1;
        } while(!etimer_expired(&s.etimer));
    }
    while(s.ticks > 0);

    // Retransmit discovery
    {
	    uip_ipaddr_t ipaddr;
	    uip_ipaddr(&ipaddr, 0, 0, 0, 0);
	    uip_sethostaddr(&ipaddr);
    }
    goto init;

#if ELITE_MODULE_DHCP_BEHAVIOR_1
dhcpc_unconfig:
#endif
    /* lease_expired: */
    dhcpc_unconfigured(&s);
    goto init;

    PT_END(&s.pt);
}
/*---------------------------------------------------------------------------*/
void
dhcpc_init(const void *mac_addr, int mac_len)
{
  uip_ipaddr_comm_t addr;

  s.mac_addr = mac_addr;
  s.mac_len  = mac_len;

  s.state = STATE_INITIAL;
  memset(&addr, 0, sizeof(addr));
  uip_ipaddr(&addr, 255,255,255,255);
  s.conn = udp_new(&addr, UIP_HTONS(DHCPC_SERVER_PORT), NULL);
  if(s.conn != NULL) {
    udp_bind(s.conn, UIP_HTONS(DHCPC_CLIENT_PORT));
  }

  dhcpc_eth_port = 0;
  dhcpc_eth_vlan = 0;

  PT_INIT(&s.pt);
}
/*---------------------------------------------------------------------------*/
void
dhcpc_appcall(process_event_t ev, void *data)
{
  if(ev == tcpip_event || ev == PROCESS_EVENT_TIMER) {
    handle_dhcp(ev, data);
  }
}
/*---------------------------------------------------------------------------*/
void
dhcpc_request(void)
{
  uip_ipaddr_t ipaddr;

  if(s.state == STATE_INITIAL) {
    uip_ipaddr(&ipaddr, 0,0,0,0);
    uip_sethostaddr(&ipaddr);
    handle_dhcp(PROCESS_EVENT_NONE, NULL);
  }
}
/*---------------------------------------------------------------------------*/
int dhcpc_dump()
{
    printf("\r\n");
    printf("dhcpc_eth_port : %d\r\n", dhcpc_eth_port);
    printf("dhcpc_eth_vlan : %d\r\n", dhcpc_eth_vlan);
    printf("\r\n");
    return 0;
}
