/* 
 * Copyright (c) 2012 Qualcomm Atheros, Inc.. 
 * All Rights Reserved. 
 * Qualcomm Atheros Confidential and Proprietary. 
 */

#ifndef _GMAC_H_
#define _GMAC_H_

#ifdef __cplusplus
extern "C" {
#endif    /* __cplusplus */


//#include "net/uip.h"
//#include "net/uip_arp.h"
#include <common.h>

#define data_swap32(value)			\
    ((((value) & 0xff000000) >> 24) |	\
     (((value) & 0x00ff0000) >>  8) |	\
     (((value) & 0x0000ff00) <<  8) |	\
     (((value) & 0x000000ff) << 24))

#define cpu_to_le32(value) data_swap32(value)
#define le32_to_cpu(value) data_swap32(value)

#ifndef round_up
#define round_up(x, y) ((((x) + ((y) - 1)) / (y)) * (y))
#endif


#define ETH_ALEN                6
//#define VIVO_P_BOOT            0x12 
#define VIVO_P_CTRL              0x12
#define VIVO_P_DATA              0x13
#define ETH_P_VIVO               0x88bd

#define DNA_HEADER_LEN  12
#define DA_SA_MAC_LEN   12



typedef struct
{
    uint8_t *buf;
    uint8_t *data;
    uint8_t tx_id;
    uint8_t next_id;
    uint16_t len;
} gmac_tx_bd_t;

typedef union
{
    uint32_t u32[(DNA_HEADER_LEN + 3) / 4];
    uint8_t u8[DNA_HEADER_LEN];
} ath_header_t;

extern ath_header_t vivo_aligned_header;
#define ath_header (vivo_aligned_header.u8)

#define VIVO_TX_BUF_SIZE		64
#define VIVO_PKT_CRC_SIZE		4


//#define VIVO_RX_BUF_SIZE      (1096*2)
#define VIVO_RX_BUF_SIZE    1600

#define DMA_ALIGN_SIZE		    8
#define DMA_ALIGN_ADDR(addr)	(round_up((uint32_t)(addr), DMA_ALIGN_SIZE))
#define DESC_BUF_SIZE (sizeof(gmac_tpd_t) * VIVO_NUM_TX_DESC + \
                       sizeof(gmac_rfd_t) * VIVO_NUM_RX_DESC + \
                       sizeof(gmac_rrd_t) * VIVO_NUM_RX_DESC + 8*3)



/* DMA Order Settings */
#define VIVO_DMA_ORD_IN   1
#define VIVO_DMA_ORD_ENH  2
#define VIVO_DMA_ORD_OUT  4

#define VIVO_DMA_REQ_128   0
#define VIVO_DMA_REQ_256   1
#define VIVO_DMA_REQ_512   2
#define VIVO_DMA_REQ_1024  3
#define VIVO_DMA_REQ_2048  4
#define VIVO_DMA_REQ_4096  5

/*Configurable*/
#define VIVO_PREAMBLE_LEN  7
#define VIVO_TPD_BURST     5
#define VIVO_RFD_BURST     8
#define VIVO_DMAR_BLOCK    VIVO_DMA_REQ_4096
#define VIVO_DMAW_BLOCK    VIVO_DMA_REQ_128
#define VIVO_DMAR_DLY_CNT  15
#define VIVO_DMAW_DLY_CNT  4
#define VIVO_RX_MTU	            1536
//#define VIVO_DMA_ORDER     VIVO_DMA_ORD_OUT



/*TPD desciptor*/
#define TPD_EOP_SHIFT	31
typedef struct gmac_tpd
{
    uint32_t data0;
    uint32_t data1;
    uint32_t buffer_address;
    uint32_t data3;
} gmac_tpd_t;

/*RRD desciptor*/
#define RRS_PKT_SIZE_MASK	0x3FFF
#define RRS_PKT_SIZE_SHIFT	0

#define RRS_RXD_UPDATED_MASK	0x0001
#define RRS_RXD_UPDATED_SHIFT	31
#define RRS_RXD_IS_VALID(data) \
	((((data) >> RRS_RXD_UPDATED_SHIFT) & RRS_RXD_UPDATED_MASK) == 1)

typedef struct gmac_rrd
{
    uint32_t data0;
    uint32_t data1_svlan;
    uint32_t data2_cvlan;
    uint32_t data3;
} gmac_rrd_t;

/* RFD desciptor */
typedef struct gmac_rfd
{
    uint32_t buffer_address;
    uint32_t buffer_address_high;
} gmac_rfd_t;


#define VIVO_NUM_TX_DESC		2
#define VIVO_NUM_RX_DESC		2
#define VIVO_MAC_SPEED_1000    2


#define GMAC_REGISTER_BASE 0xb8110000

#define gmac_reg_rd(r)	({		\
	music_reg_rd(GMAC_REGISTER_BASE + (r));	\
})

#define gmac_reg_wr(r, v)	({	\
	music_reg_wr((GMAC_REGISTER_BASE + (r)), v);	\
})

#define SWITCH_REGISTER_BASE 0xb8800000
#define switch_reg_rd(r)	({		\
	music_reg_rd(SWITCH_REGISTER_BASE + (r));	\
})

#define switch_reg_wr(r, v)	({	\
	music_reg_wr((SWITCH_REGISTER_BASE + (r)), v);	\
})

/* register content */

/* Selene Master Control Register */
#define REGISTER_MASTER_CTRL			0x1400
#define MASTER_CONTROL_SOFT_RST        0x1

/* MAC Control Register  */
#define REGISTER_MAC_CTRL        	0x1480
#define MAC_CONTROL_TX_EN			        0x1
#define MAC_CONTROL_RX_EN			        0x2
#define MAC_CONTROL_TX_FLOW		        0x4
#define MAC_CONTROL_RX_FLOW         	    0x8
#define MAC_CONTROL_LOOPBACK          	    0x10
#define MAC_CONTROL_DUPLX              	0x20
#define MAC_CONTROL_ADD_CRC            	0x40
#define MAC_CONTROL_PAD                	0x80
#define MAC_CONTROL_LENCHK             	0x100
#define MAC_CONTROL_HUGE_EN            	0x200
#define MAC_CONTROL_PRMLEN_SHIFT       	10
#define MAC_CONTROL_PRMLEN_MASK        	0xf
#define MAC_CONTROL_RMV_VLAN           	0x4000
#define MAC_CONTROL_PROMIS_EN          	0x8000
#define MAC_CONTROL_TX_PAUSE           	0x10000
#define MAC_CONTROL_SCNT               	0x20000
#define MAC_CONTROL_SRST_TX            	0x40000
#define MAC_CONTROL_TX_SIMURST         	0x80000
#define MAC_CONTROL_SPEED_SHIFT        	20
#define MAC_CONTROL_SPEED_MASK         	0x3
#define MAC_CONTROL_DBG_TX_BKPRESURE   	0x400000
#define MAC_CONTROL_TX_HUGE            	0x800000
#define MAC_CONTROL_RX_CHKSUM_EN       	0x1000000
#define MAC_CONTROL_MC_ALL_EN          	0x2000000
#define MAC_CONTROL_BC_EN              	0x4000000
#define MAC_CONTROL_DBG                	0x8000000
#define MAC_CONTROL_SINGLE_PAUSE_EN	    0x10000000
#define MAC_CONTROL_HASH_ALG_CRC32		    0x20000000
#define MAC_CONTROL_SPEED_MODE_SW		    0x40000000


#define REGISTER_MAC_STA_ADDR		        0x1488

#define REGISTER_MTU                     	0x149c

/*
 * Load Ptr Register 
 */
#define REGISTER_LOAD_PTR                	0x1534

#define REGISTER_RX_BASE_ADDR_HI		0x1540
#define REGISTER_TX_BASE_ADDR_HI		0x1544
#define REGISTER_SMAILBOX_BASE_ADDR_HI		0x1548
#define REGISTER_SMAILBOX_BASE_ADDR_LO		0x154C
#define REGISTER_RFD0_HEAD_ADDR_LO		0x1550
#define REGISTER_RFD1_HEAD_ADDR_LO		0x1554
#define REGISTER_RFD2_HEAD_ADDR_LO		0x1558
#define REGISTER_RFD3_HEAD_ADDR_LO		0x155C
#define REGISTER_RFD_RING_SIZE		0x1560
#define RFD_RING_SIZE_MASK		0x0FFF
#define REGISTER_RX_BUF_SIZE			0x1564
#define RX_BUF_SIZE_MASK		0xFFFF
#define REGISTER_RRD0_HEAD_ADDR_LO		0x1568
#define REGISTER_RRD1_HEAD_ADDR_LO		0x156C
#define REGISTER_RRD2_HEAD_ADDR_LO		0x1570
#define REGISTER_RRD3_HEAD_ADDR_LO		0x1574
#define REGISTER_RRD_RING_SIZE		0x1578
#define RRD_RING_SIZE_MASK		0x0FFF
#define REGISTER_HTPD_HEAD_ADDR_LO		0x157C
#define REGISTER_NTPD_HEAD_ADDR_LO		0x1580
#define REGISTER_TPD_RING_SIZE		0x1584
#define TPD_RING_SIZE_MASK		0xFFFF
#define REGISTER_CMAILBOX_BASE_ADDR_LO		0x1588

/* TXQ Control Register */
#define REGISTER_TXQ_CTRL                	0x1590
#define	TXQ_NUM_TPD_BURST_MASK     	0xF
#define TXQ_NUM_TPD_BURST_SHIFT    	0
#define TXQ_CONTROL_IP_OPTION_EN		0x10
#define TXQ_CONTROL_EN                     0x20
#define TXQ_CONTROL_ENH_MODE               0x40
#define TXQ_CONTROL_LS_8023_EN		0x80
#define TXQ_TXF_BURST_NUM_SHIFT    	16
#define TXQ_TXF_BURST_NUM_MASK     	0xFFFF

/* Jumbo packet Threshold for task offload */
#define REGISTER_TX_TSO_OFFLOAD_THRESH	0x1594

/* RXQ Control Register */
#define REGISTER_RXQ_CTRL                	0x15A0
#define RXQ_RFD_BURST_NUM_MASK		0x003F
#define RXQ_RFD_BURST_NUM_SHIFT		20
#define RXQ_CONTROL_EN			0x80000000

/* DMA Engine Control Register */
#define REGISTER_DMA_CTRL                	0x15C0
#define DMA_CONTROL_DMAR_IN_ORDER          0x1
#define DMA_CONTROL_DMAR_ENH_ORDER         0x2
#define DMA_CONTROL_DMAR_OUT_ORDER         0x4
#define DMA_CONTROL_RCB_VALUE              0x8
#define DMA_CONTROL_DMAR_BURST_LEN_MASK    0x0007
#define DMA_CONTROL_DMAR_BURST_LEN_SHIFT   4
#define DMA_CONTROL_DMAW_BURST_LEN_MASK    0x0007
#define DMA_CONTROL_DMAW_BURST_LEN_SHIFT   7
#define DMA_CONTROL_DMAR_REQ_PRI           0x400
#define DMA_CONTROL_DMAR_DLY_CNT_MASK      0x001F
#define DMA_CONTROL_DMAR_DLY_CNT_SHIFT     11
#define DMA_CONTROL_DMAW_DLY_CNT_MASK      0x000F
#define DMA_CONTROL_DMAW_DLY_CNT_SHIFT     16

/* Mail box */
#define MAILBOX_RFDX_PROD_IDX_MASK		0xFFFF
#define REGISTER_MAILBOX_RFD0_PROD_IDX		0x15E0
#define REGISTER_MAILBOX_RFD1_PROD_IDX		0x15E4
#define REGISTER_MAILBOX_RFD2_PROD_IDX		0x15E8
#define REGISTER_MAILBOX_RFD3_PROD_IDX		0x15EC

#define MAILBOX_PRIO_PROD_IDX_MASK		0xFFFF
#define REGISTER_MAILBOX_PRIO_PROD_IDX		0x15F0
#define MAILBOX_HTPD_PROD_IDX_SHIFT		0
#define MAILBOX_NTPD_PROD_IDX_SHIFT		16

#define MAILBOX_PRIO_CONS_IDX_MASK		0xFFFF
#define REGISTER_MAILBOX_PRIO_CONS_IDX		0x15F4
#define MAILBOX_HTPD_CONS_IDX_SHIFT		0
#define MAILBOX_NTPD_CONS_IDX_SHIFT		16

/*Endian Swap*/
#define REGISTER_AXI_MAST_CTRL    		0x1610
#define AXI_MAST_ENDIAN_SWAP_EN     0x0008

/*Atheros Header*/
#define REGISTER_ATHEROS_HEADER_CTRL		0x1620
#define ATHEROS_HEADER_EN		        0x0001
#define ATHEROS_HEADER_CNT_EN		    0x0002

void set_ath_header_to_zero(void);
void set_cpu_tpid(uint16_t tpid_value);


#ifdef __cplusplus
}
#endif    /* __cplusplus */

#endif /*_GMAC_API_H_*/
