/* 
 * Copyright (c) 2012 Qualcomm Atheros, Inc.. 
 * All Rights Reserved. 
 * Qualcomm Atheros Confidential and Proprietary. 
 */

#ifndef __MUSIC_UPGRADE_H__
#define __MUSIC_UPGRADE_H__

#ifdef __cplusplus
extern "C" {
#endif    /* __cplusplus */


#include "music_spi.h"

#define SPI_FLASH_BASE      0xb9000000
#if 0
#define SPI_FLASH_SIZE      0x400000
#define IMAGE_A_OFFSET       0x40000
#define IMAGE_B_OFFSET       0x100000
#define IMAGE_RUNNING_OFFSET 0x1c0000
#define CONTROL_BLOCK_OFFSET 0x3c000
#define BOARD_INFO_OFFSET    0x3c0000
#define CONFIG_FILE_OFFSET   0x3c4000
#else
extern uint32_t global_flash_size;
extern uint32_t global_image_a_offset;
extern uint32_t global_image_b_offset;
extern uint32_t global_running_offset;
extern uint32_t global_ctrl_block_offset;
extern uint32_t global_boot_cfg_offset;
extern uint32_t global_user_cfg_offset;

#define SPI_FLASH_SIZE      global_flash_size
#define IMAGE_A_OFFSET       global_image_a_offset
#define IMAGE_B_OFFSET       global_image_b_offset
#define IMAGE_RUNNING_OFFSET global_running_offset
#define CONTROL_BLOCK_OFFSET global_ctrl_block_offset
#define BOARD_INFO_OFFSET    global_boot_cfg_offset
#define CONFIG_FILE_OFFSET   global_user_cfg_offset
#endif
#define IMAGE_A_BASE        (SPI_FLASH_BASE + IMAGE_A_OFFSET)
#define IMAGE_B_BASE        (SPI_FLASH_BASE + IMAGE_B_OFFSET)
#define CONTROL_BLOCK_BASE  (SPI_FLASH_BASE + CONTROL_BLOCK_OFFSET)
#define BOARD_INFO_BASE     (SPI_FLASH_BASE + BOARD_INFO_OFFSET)
#define CONFIG_FILE_BASE    (SPI_FLASH_BASE + CONFIG_FILE_OFFSET)


#define CONTROL_BLOCK_BACKUP1_ADDR (CONTROL_BLOCK_BASE + 0x2000)
#define CONTROL_BLOCK_BACKUP2_ADDR (CONTROL_BLOCK_BASE + 0x3000)

#define LENGTH_32BYTES      32
#define LENGTH_16BYTES      16
#define LENGTH_4BYTES       4

#define OEM_ID_LEN          LENGTH_16BYTES
#define PRODUCT_LEN         LENGTH_16BYTES
#define BOARD_VERSION_LEN   LENGTH_4BYTES
#define SERIAL_NUM_LEN      LENGTH_32BYTES
#define MAC_ADDRESS_LEN     6

#define MAGIC_LEN           LENGTH_16BYTES
#define IMAGE_VERSION_LEN   LENGTH_32BYTES
#define BUILD_TIME_LEN      LENGTH_32BYTES

#define IMAGE_HEADER_LEN        (MAGIC_LEN + IMAGE_VERSION_LEN + BUILD_TIME_LEN + \
OEM_ID_LEN + PRODUCT_LEN + 4 + 4 + 2)
#define IMAGE_HEADER_PAD_LEN    (255 - IMAGE_HEADER_LEN)
#define IMAGE_HEADER_SIZE       256
#define IMAGE_MAX_LENGTH    (0x180000 - IMAGE_HEADER_SIZE)

#define UPGRADE_SET  	    0xabcde55a  //0x00534554
#define UPGRADE_NSET 	    0x24875cde  //0x4e534554
#define IMAGE_A		        0xbeef3896  //0x494d4741
#define IMAGE_B		        0xcafe2012  //0x494d4742
#define MAX_LOAD_COUNT	    3
#define IMAGE_VALID         0xdead3897
#define IMAGE_UNVALID       0xfaebeef8
#define IMAGE_DECOMP        0xdeadacb9
#define IMAGE_NOT_DECOMP    0xfaedbeef

#define CONTROL_BLOCK_LEN   (4*9)
#define CONTROL_BLOCK_PAD_LEN   (0x100 - CONTROL_BLOCK_LEN)


typedef struct
{
    char oem_id[OEM_ID_LEN];
    char product_id[PRODUCT_LEN];
    char board_version[BOARD_VERSION_LEN];
    char serial_num[SERIAL_NUM_LEN];
    char mac_address[MAC_ADDRESS_LEN];
} board_info_t;

typedef struct
{
    char magic[MAGIC_LEN];
    char image_version[IMAGE_VERSION_LEN];
    char build_time[BUILD_TIME_LEN];    //2012.12.21 20:38:12
    char oem_id[OEM_ID_LEN];
    char product_id[PRODUCT_LEN];
    uint32_t image_header_length;
    uint32_t image_length;
    char image_checksum;
    char pad[IMAGE_HEADER_PAD_LEN];
    char header_checksum;
} image_header_t;

typedef struct
{
    uint32_t upgrade_flag;
    uint32_t select_flag;
    uint32_t image_is_decompressed;
    uint32_t will_select_flag;
    uint32_t imageA_loader_count;
    uint32_t imageA_valid_flag;
    uint32_t imageB_loader_count;
    uint32_t imageB_valid_flag;
    uint32_t crc_32bit;
    char pad[CONTROL_BLOCK_PAD_LEN];
} ctrl_info_t;

uint32_t board_is_qfndemo();
uint32_t image_upgrade_identification_check();
void show_image_info();
uint32_t image_header_check(image_header_t * image_header);
uint32_t current_running_image();
uint32_t image_crc_check(uint32_t which_image);
uint32_t control_block_crc_compute(ctrl_info_t * control_info);
uint32_t control_block_copy(ctrl_info_t * dest, ctrl_info_t * source);
uint32_t debug_set_ctrl_block(char *image_name);
uint32_t control_block_backup_save();
uint32_t debug_crash_backup1_block();
uint32_t debug_crash_backup2_block();
uint32_t debug_crash_ctrl_block();
uint32_t debug_upgrade_image(uint32_t source, uint32_t dest);
uint32_t upgrade_info_dump();
uint32_t switchover_image_operation();
uint32_t switchover_image_check();
uint32_t whether_can_upgrade_image();
uint32_t current_select_image();
int image_get_version(char *version);
int image_get_version_2(uint32_t image, char *version);
int product_get_serial_number(char *serial);
int product_get_mac(char *mac);
uint32_t elite_running_seconds();


#ifdef __cplusplus
}
#endif    /* __cplusplus */

#endif                          /* __MUSIC_UPGRADE_H__ */
