/* * Copyright (c) 2012 Qualcomm Atheros, Inc.
 * All Rights Reserved. 
 *Qualcomm Atheros Confidential and Proprietary. 
 *Notifications and licenses are retained for attribution purposes only */
/*********************************************************************/
/*                                                                   */
/*  This Program Written by Paul Edwards.                            */
/*  Released to the Public Domain                                    */
/*                                                                   */
/*********************************************************************/
/*********************************************************************/
/*                                                                   */
/*  zmcore - core zmodem routines                                    */
/*                                                                   */
/*********************************************************************/

#ifndef __ZMODEM_H__
#define __ZMODEM_H__

#ifdef __cplusplus
extern "C" {
#endif    /* __cplusplus */


#define ZPAD    0x2a
#define ZDLE    0x18
#define ZBIN    0x41
#define ZHEX    0x42
#define ZBIN32  0x43
#define ZRQINIT 0x00
#define ZRINIT  0x01
#define CANFDX  0x01
#define CANOVIO 0x02
#define CANFC32 0x20
#define ZACK    0x03
#define ZFILE   0x04
#define ZSKIP   0x05
#define ZCRCW   0x6b
#define ZRPOS   0x09
#define ZDATA   0x0a
#define ZCRCG   0x69
#define ZEOF    0x0b
#define ZFIN    0x08
#define ZRUB0   0x6c            /* 'l' - escape for 0x7f */
#define ZRUB1   0x6d            /* 'm' - escape for 0xff */

#define ZCBIN   0x01
#define ZCRCE   0x68
#define ZCAN    0xfe            /* not implemented properly yet */

#define ZMODEM_INIT     1       /* Remote didn't accept my ZRINIT             */
#define ZMODEM_POS      2       /* Remote failed to respond to my ZRPOS       */
#define ZMODEM_ZDATA    4       /* Remote failed to sync to correct position  */
#define ZMODEM_CRCXM    8       /* CRCXM                      */
#define ZMODEM_LONGSP   0x10    /* Error - subpacket too long */
#define ZMODEM_CRC32    0x20    /* CRC32 error                */
#define ZMODEM_FILEDATA 0x40    /* File Data is out of spec   */
#define ZMODEM_BADHEX   0x80    /* Bad hex character received */
#define ZMODEM_TIMEOUT  0x100   /* Timeout                    */

#define ZMODEM_RECEIVE_BASEADDR ZM_DOWNLOAD_ADDR
#define ZMODEM_BUFFER_SIZE      0x2000000

#define CR             015
#define LF             012

#define ZMCORE_MAXTX 1024
#define FILENAME_MAX 128


typedef struct
{
    uint8_t filename[FILENAME_MAX];
    uint8_t *fileinfo;
    int32_t bytes;
    int32_t bufsize;
    int32_t maxtx;
    uint32_t good_offset;
    int32_t filesize;           /* Size of file in bytes, 0 if unspecified. */
    int32_t filetime;           /* As defined by the zmodem protocol: the number
                                   f seconds since 1 Jan 1970, 0 for unspecified. */
    int32_t filemode;           /* Permissions on file.  Defaults to 0644. */
    int32_t got_special;
    int32_t got_header;
    int32_t moredata;
    int32_t skip;
    int32_t data_start_offset;  /* zmodem will transfer file's info before
                                   its content, we need sparate it from data */
    uint8_t frame_type;
    uint8_t header_type;
    uint8_t pad0[2];
    uint8_t header_data[4];
    uint8_t *main_buf;
    uint8_t *buf_pos;
    uint8_t *buf_top;
    int32_t ch;
} ZMCORE;

#define zmodem_error_set(x)            do {                         \
                                        romboot_errorno|=(x);       \
                                        /* *(int32_t *) 0xbd007000 |= romboot_errorno; */\
                                            }while(0)
#define ALLOK                           (0 == romboot_errorno)
#define zmodem_error_flush()        do {                            \
                                        romboot_errorno = 0;        \
                                        }while(0)
#define zmodem_error_compare(x)         ((romboot_errorno & (x)) == (x))

#define zmodem_error_clear()            zmodem_error_flush()


extern int32_t romboot_errorno;
extern void romboot_zmodem_init(ZMCORE * zmcore);
extern int32_t romboot_zmodem_receive(ZMCORE * zmcore);
extern int32_t romboot_zmodem_send(ZMCORE * zmcore);
extern void romboot_zmodem_term(ZMCORE * zmcore);
extern void zmodem_main();
extern void extmodem_getblock(ZMCORE * zmcore, uint8_t * buf);
extern void extmodem_writeblock(ZMCORE * zmcore, uint8_t * buf, int32_t max);


#ifdef __cplusplus
}
#endif    /* __cplusplus */

#endif
