/* 
 * Copyright (c) 2012 Qualcomm Atheros, Inc.. 
 * All Rights Reserved. 
 * Qualcomm Atheros Confidential and Proprietary. 
 */

#include <common.h>
#include "serial.h"

#define REG_EFUSE_HIGH_VAL                      0x10
#define REG_SWITCH_EFUSE                        REG_EFUSE_HIGH_VAL
#define SWITCH_EFUSE_CPU_SPEED_MODE_MASK        0x00000007
#define REG_SWITCH_BASE                         0x18800000

int soc_bus_freq_get(void)
{
    /*000=200MHz, 001=300MHz, 010=400MHz, 011=480MHz, 
       100=533MHz, 101=600MHz. 110 and 111 Reserved. */

    uint32_t cpu_speed_mhz[8] = { 200, 300, 400, 480, 533, 600, 300, 300 };
    uint32_t reg_val = music_reg_rd(REG_SWITCH_BASE + REG_SWITCH_EFUSE) & SWITCH_EFUSE_CPU_SPEED_MODE_MASK;

    return ((cpu_speed_mhz[reg_val] * 1000 * 1000) / 2);
}

void serial_init(void)
{
    int div = 0;
    int val = 0;

    //div = SERIAL_CLOCK_FREQ / (16 * CONFIG_BAUDRATE);
    div = soc_bus_freq_get() / (16 * CONFIG_BAUDRATE);
    /* 
     * Enable FIFO mode 
     */
    /* enable FIFO */
    val = UART16550_READ(OFS_FIFO_CONTROL);
    UART16550_WRITE(OFS_FIFO_CONTROL, (val | 0x01));

    /* reset transmit FIFO and recieve FIFO */
    val = UART16550_READ(OFS_FIFO_CONTROL);
    UART16550_WRITE(OFS_FIFO_CONTROL, (val | 0x06));
    /*set DIAB bit */
    UART16550_WRITE(OFS_LINE_CONTROL, 0x80);

    /* set divisor */
    UART16550_WRITE(OFS_DIVISOR_LSB, (div & 0xff));
    UART16550_WRITE(OFS_DIVISOR_MSB, ((div >> 8) & 0xff));

    /* clear DIAB bit */
    UART16550_WRITE(OFS_LINE_CONTROL, 0x00);

    /* set data format */
    UART16550_WRITE(OFS_DATA_FORMAT, 0x3);

    UART16550_WRITE(OFS_INTR_ENABLE, 0);

    //printf("\nserial_init done\n");
}

void serial_baudrate_set(int baudrate)
{
    int div = 0;
    int val = 0;

    //div = SERIAL_CLOCK_FREQ / (16 * CONFIG_BAUDRATE);
    div = soc_bus_freq_get() / (16 * baudrate);
    /* 
     * Enable FIFO mode 
     */
    /* enable FIFO */
    val = UART16550_READ(OFS_FIFO_CONTROL);
    UART16550_WRITE(OFS_FIFO_CONTROL, (val | 0x01));

    /* reset transmit FIFO and recieve FIFO */
    val = UART16550_READ(OFS_FIFO_CONTROL);
    UART16550_WRITE(OFS_FIFO_CONTROL, (val | 0x06));
    /*set DIAB bit */
    UART16550_WRITE(OFS_LINE_CONTROL, 0x80);

    /* set divisor */
    UART16550_WRITE(OFS_DIVISOR_LSB, (div & 0xff));
    UART16550_WRITE(OFS_DIVISOR_MSB, ((div >> 8) & 0xff));

    /* clear DIAB bit */
    UART16550_WRITE(OFS_LINE_CONTROL, 0x00);

    /* set data format */
    UART16550_WRITE(OFS_DATA_FORMAT, 0x3);

    UART16550_WRITE(OFS_INTR_ENABLE, 0);

    //printf("\nserial_init done\n");
}

int serial_tstc(void)
{
    return (UART16550_READ(OFS_LINE_STATUS) & 0x1);
}

uint8_t serial_getc(void)
{
    while (!serial_tstc());

    return UART16550_READ(OFS_RCV_BUFFER);
}

void serial_putc(uint8_t byte)
{
    if (byte == '\n')
        serial_putc('\r');

    while (((UART16550_READ(OFS_LINE_STATUS)) & 0x20) == 0x0);
    UART16550_WRITE(OFS_SEND_BUFFER, byte);
}

void serial_setbrg(void)
{
}

void serial_puts(const char *s)
{
    while (*s)
    {
        serial_putc(*s++);
    }
}
