/* 
 * Copyright (c) 2012 Qualcomm Atheros, Inc.. 
 * All Rights Reserved. 
 * Qualcomm Atheros Confidential and Proprietary. 
 */

#include "common.h"
#include "spiflash.h"


#define ARS124_AHB_CLOCK 600000000/2
#define ARS124_SPI_READDATA_CLOCK 12500000

void music_spi_write_enable()
{
    /*enable function reister other than 
       enable reads to the SPI space */
    music_reg_wr(MUSIC_SPI_FS, 1);

    /*low select: Chip Select 0 bit to be output */
    music_reg_wr(MUSIC_SPI_WRITE, MUSIC_SPI_CS_DIS);

    /*write enable */
    music_spi_bit_banger(MUSIC_SPI_CMD_WREN);

    /*??? */
    music_spi_go();
}

void music_spi_poll()
{
    int rd = 1;

    do
    {
        music_reg_wr(MUSIC_SPI_WRITE, MUSIC_SPI_CS_DIS);
        music_spi_bit_banger(MUSIC_SPI_CMD_RD_STATUS);
        music_spi_delay_8();
        rd = (music_reg_rd(MUSIC_SPI_RD_STATUS) & 1);
    }
    while (rd);
}

void music_spi_sector_erase(uint32_t addr)
{
    music_spi_write_enable();
    music_spi_bit_banger(MUSIC_SPI_CMD_SECTOR_ERASE);
    music_spi_send_addr(addr);
    music_spi_go();
    music_spi_poll();
}

void music_spi_write_page(uint32_t addr, uint8_t * data, int len)
{
    int i;
    uint8_t ch;

    music_spi_write_enable();
    /*pp(page program) cmd */
    music_spi_bit_banger(MUSIC_SPI_CMD_PAGE_PROG);
    /*3 bytes address */
    music_spi_send_addr(addr);

    for (i = 0; i < len; i++)
    {
        ch = *(data + (i / 4) * 4 + 3 - (i % 4));
        /*at least 1 byte data */
        music_spi_bit_banger(ch);
    }

    music_spi_go();
    music_spi_poll();
}

void music_spi_bulk_erase()
{
    music_spi_write_enable();
    music_spi_bit_banger(MUSIC_SPI_CMD_BULK_ERASE);
    music_spi_go();
    music_spi_poll();
}

void music_spi_read(uint32_t addr, uint32_t * data, int len)
{
    int i = 0, val;

    /*enable reads to the SPI space */
    music_reg_wr(MUSIC_SPI_FS, 0);

    for (i = 0; i < len; i++)
    {
        val = music_reg_rd(MUSIC_SPI_READ + addr + 4 * i);
        *(data + i) = val;
    }
}

/*
 * 1. Calculate number of sectors this write spans
 * 2. Erase 'em
 * 3. call page programming for every 256 bytes
 */
int music_spi_write(uint8_t * source, uint32_t addr, uint32_t len)
{
    int total = 0, len_this_lp, bytes_this_page, nsect, i;
    uint32_t dst = 0;
    uint8_t *src;
    int start_sect = 0;

    nsect = len / MUSIC_SPI_SECTOR_SIZE;
    if (len % MUSIC_SPI_SECTOR_SIZE)
    {
        nsect++;
    }
    start_sect = addr / MUSIC_SPI_SECTOR_SIZE;

    for (i = start_sect; i < start_sect + nsect; i++)
    {
        music_spi_sector_erase(i * MUSIC_SPI_SECTOR_SIZE);
    }

    while (total < len)
    {
        src = source + total;
        dst = addr + total;
        bytes_this_page = MUSIC_SPI_PAGE_SIZE - (addr % MUSIC_SPI_PAGE_SIZE);
        len_this_lp = ((len - total) > bytes_this_page) ? bytes_this_page : (len - total);
        if (total % (MUSIC_SPI_PAGE_SIZE * 0x80) == 0)
            printf("src:%x dst:%x len_this_lp:%x total:%x len:%x\n", src, dst | 0xb9000000, len_this_lp, total, len);
        music_spi_write_page(dst, src, len_this_lp);
        total += len_this_lp;
    }

    return 0;
}

void music_spi_rdid()
{
    music_spi_bit_banger(MUSIC_SPI_CMD_RDID);
    music_spi_delay_8();
    music_spi_delay_8();
    music_spi_delay_8();
    music_spi_delay_8();

    music_spi_go();
    //printf("music_spi_init done RDID:%x\n", music_reg_rd(MUSIC_SPI_RD_STATUS) );  
}

void music_spi_init()
{
    int spi_clock_div = 0;
    /*Per bit output enable. Setting a bit to 1
       will enable the output driver. Setting a bit
       to 0 will enable its use as an input. */
    //music_reg_wr(MUSIC_GPIO_OE, 0xcff);
    //music_reg_wr(MUSIC_SPI_DSEL, 0);

    /*When write a non-zero value to this register selects "GPIO" mode 
       for the SPI controller. In this mode, the other sw registers are visible.
       When a zero value is written to this register, reads to the SPI space are
       translated in reads of the external SPI flash devic */
    music_reg_wr(MUSIC_SPI_FS, 1);
    /*  clear remap_disable bit 1=16m 0=4m
     *  set spi-flash read speed
     *  speed = ahb_clock / (2 *(div + 1))
     */
    spi_clock_div = (ARS124_AHB_CLOCK / ARS124_SPI_READDATA_CLOCK * 2) - 1;
    if (!spi_clock_div)
    {
        /* spi_clock must litter than quarter of AHB clock */
        spi_clock_div = 1;

    }

    spi_clock_div |= 0x40;

    music_reg_wr(MUSIC_SPI_CLOCK, spi_clock_div);

    music_spi_rdid();

    //printf("music_spi_init done\n");  
}
