/* 
 * Copyright (c) 2012 Qualcomm Atheros, Inc.. 
 * All Rights Reserved. 
 * Qualcomm Atheros Confidential and Proprietary. 
 */

#include "mipsregs.h"
#include "music_spi.h"
#include "music_upgrade.h"
#include "mem_api.h"
#include "version.h"

#define LINUX_ARGV_START   0x80001000
#define LINUX_LINK_START   0x80000000   /*also download address */
#define ELITE_ENTRY_POINT  0xc8004000
#define ELITE_VIRTUAL_ADDRESS   0xc8000000
#define ELITE_SP_VIRTUAL_ADDRESS   0xdd000000
//#define IMAGEA_ADDR     0x19100000
#define SPI_FLASH_PHY_BASE  0x19000000
#define IMAGEA_ADDR     0x19010000
#define IMAGEB_ADDR     0x19280000
#define VIVO_SRAM_BLOCK0	0x1d000000
#define VIVO_SRAM_BLOCK1	0x18150000
#define TLB_256K_PAGE_SIZE      0x40000 /* 256K bytes */
#define TLB_256K_PAGE_MASK      0x7e000
#define TLB_64K_PAGE_SIZE       0x10000 /* 64K bytes */
#define TLB_64K_PAGE_MASK       0x1e000
#define TLB_16K_PAGE_SIZE       0x4000  /* 16K bytes */
#define TLB_16K_PAGE_MASK       0x6000
#define BOARD_TYPE_ADDR         0xbd007000
//#define BOARD_TYPE              NOT_QFN_DEMON_BOARD
#define BOARD_TYPE              QFN_DEMON_BOARD
#define QFN_DEMON_BOARD         0xbeafcafe
#define NOT_QFN_DEMON_BOARD     0x00000000

#define MUSIC_SPI_FS           0x19000000   /*SPI FUCTION SELECT REG */
#define MUSIC_SPI_CLOCK        0x19000004   /*SPI  CONTROL REG */

#define TLB_FLAG_D	(1<<2)
#define TLB_FLAG_V	(1<<1)
#define TLB_FLAG_G	(1<<0)
#define UNCACHED_FLAG	(2<<3)
#define CACHABLE_FLAG	(3<<3)

#define tlb_write_indexed()                                     \
	__asm__ __volatile__(                                   \
		".set noreorder\n\t"                            \
		"tlbwi\n\t"                                     \
".set reorder")

#define ZM_USER_BREAK_INVALID_CONF_KEY      0x12    /* ^R - Control R */

int printf(const char *fmt, ...);
int serial_init(void);
int serial_tstc(void);
unsigned int serial_getc(void);
void udelay(uint32_t usec);
void music_spi_init();
void watchdog_init(void);
void watchdog_start(void);
uint32_t control_block_backup_check();
uint32_t control_block_verification(ctrl_info_t * control_info);

void (*start_elite) ();
void (*spi_init_function_p) (void);

extern board_partition_info_t global_partition_info;


char *strcpy(char *dest, const char *src)
{
    char *tmp = dest;

    while ((*dest++ = *src++) != '\0')
        /* nothing */ ;
    return tmp;
}

int strlen(const char *s)
{
    const char *sc;

    for (sc = s; *sc != '\0'; ++sc)
        /* nothing */ ;
    return sc - s;
}

void write_one_tlb(int index, unsigned long pagemask, unsigned long hi, unsigned long low0, unsigned long low1)
{
    write_32bit_cp0_register(CP0_ENTRYLO0, low0);
    write_32bit_cp0_register(CP0_PAGEMASK, pagemask);
    write_32bit_cp0_register(CP0_ENTRYLO1, low1);
    write_32bit_cp0_register(CP0_ENTRYHI, hi);
    write_32bit_cp0_register(CP0_INDEX, index);
    tlb_write_indexed();
}

uint32_t music_tlb_256k_setting(uint32_t start_phy_addr, uint32_t start_va_addr, uint32_t mapping_size,
                                uint32_t cache_attr, uint32_t start_tlb_seq_no)
{
    uint32_t image_physical_base = start_phy_addr;
    uint32_t image_va_base = start_va_addr;
    uint32_t tlb_num = mapping_size / TLB_256K_PAGE_SIZE / 2, i = 0;
    for (i = 0; i < tlb_num; i++)
    {
        write_one_tlb(start_tlb_seq_no, /* index */
                      TLB_256K_PAGE_MASK,   /* Pagemask, 256 KB pages */
                      image_va_base,    /* Hi */
                      (cache_attr | (image_physical_base) >> 6),    /* Lo0 */
                      (cache_attr | (image_physical_base + TLB_256K_PAGE_SIZE) >> 6));

        image_physical_base += TLB_256K_PAGE_SIZE * 2;
        image_va_base += TLB_256K_PAGE_SIZE * 2;
        start_tlb_seq_no++;
    }
    return start_tlb_seq_no;
}

void music_boot_elite(void)
{

    start_elite = (void (*)())ELITE_ENTRY_POINT;
    start_elite();
}

int main()
{
    unsigned long image_physical_base = 0;
    unsigned long cache_attr = 0;
    int abort_key = 0;
    ctrl_info_t *control_block_p, control_block;
    unsigned long windows_size, windows_mask, mapping_size;
    uint32_t tlb_seq = 0;


    serial_init();
    board_get_partition_info();
    partition_info_show();

    image_physical_base = IMAGEA_ADDR;

    cache_attr = (CACHABLE_FLAG | TLB_FLAG_D | TLB_FLAG_V | TLB_FLAG_G);

    image_physical_base = 0x19080000;
    image_physical_base = SPI_FLASH_PHY_BASE + global_partition_info.running_image.segment_addr;
    mapping_size = global_partition_info.running_image.segment_size;
    tlb_seq = 3;
    if (mapping_size % TLB_256K_PAGE_SIZE)
        mapping_size += TLB_256K_PAGE_SIZE - mapping_size % TLB_256K_PAGE_SIZE;
    tlb_seq = music_tlb_256k_setting(image_physical_base, 0xc0000000, mapping_size, cache_attr, tlb_seq);

    image_physical_base = IMAGEA_ADDR;

    image_physical_base = 0x19000000;
    write_one_tlb(0,            /* index */
                  TLB_64K_PAGE_MASK,    /* Pagemask, 64 KB pages */
                  ELITE_VIRTUAL_ADDRESS,    /* Hi */
                  (cache_attr | (image_physical_base) >> 6),    /* Lo0 */
                  (cache_attr | (image_physical_base + TLB_64K_PAGE_SIZE) >> 6));   /* Lo1 */

    cache_attr = (UNCACHED_FLAG | TLB_FLAG_D | TLB_FLAG_V | TLB_FLAG_G);
    write_one_tlb(1,            /* index */
                  TLB_16K_PAGE_MASK,    /* Pagemask, 16 KB pages */
                  (ELITE_SP_VIRTUAL_ADDRESS),   /* Hi */
                  (cache_attr | VIVO_SRAM_BLOCK0 >> 6), /* Lo0 */
                  (cache_attr | (VIVO_SRAM_BLOCK0 + TLB_16K_PAGE_SIZE) >> 6));  /* Lo1 */
    write_one_tlb(2,            /* index */
                  TLB_16K_PAGE_MASK,    /* Pagemask, 16 KB pages */
                  (ELITE_SP_VIRTUAL_ADDRESS + TLB_16K_PAGE_SIZE * 2),   /* Hi */
                  (cache_attr | VIVO_SRAM_BLOCK1 >> 6), /* Lo0 */
                  (cache_attr | (VIVO_SRAM_BLOCK1 + TLB_16K_PAGE_SIZE) >> 6));  /* Lo1 */

    printf("start boot elite at physical address:0x%x!\n\r", (int)image_physical_base);
    music_boot_elite();
    return 0;
}
