/* 
 * Copyright (c) 2012 Qualcomm Atheros, Inc.. 
 * All Rights Reserved. 
 * Qualcomm Atheros Confidential and Proprietary. 
 */

#include "mipsregs.h"
#include "music_spi.h"
#include "music_upgrade.h"
#include "mem_api.h"
#include "version.h"
#include "music_gpio.h"

#define ELITE_ENTRY_POINT  0xc0000100
#define VIVO_RX_BUF_SIZE    1600
#define ZM_USER_BREAK_INVALID_CONF_KEY      0x03    /* ^C - Control C */
#define LOADER_PROMPT "loader>"
#define MAX_SERIAL_BUF_LEN 32

uint32_t serial_line_buf_len = 0;
uint8_t serial_buf[MAX_SERIAL_BUF_LEN + 1] = { 0 };
static uint8_t packet_buf[VIVO_RX_BUF_SIZE];

int printf(const char *fmt, ...);
int serial_init(void);
int serial_tstc(void);
unsigned int serial_getc(void);
void udelay(uint32_t usec);
void music_spi_init();
void watchdog_init(void);
void watchdog_start(void);
uint32_t control_block_backup_check();
uint32_t control_block_verification(ctrl_info_t * control_info);
int uncompress_image(uint32_t decompress_addr, uint32_t compress_addr);
void agent_pkt_handle(char *pkt, int len);


void (*start_elite) ();
void (*spi_init_function_p) (void);

char *strcpy(char *dest, const char *src)
{
    char *tmp = dest;

    while ((*dest++ = *src++) != '\0')
        /* nothing */ ;
    return tmp;
}

int strlen(const char *s)
{
    const char *sc;

    for (sc = s; *sc != '\0'; ++sc)
        /* nothing */ ;
    return sc - s;
}

void gpio_config_init(void)
{
    /* config GPIO to input pin */
    gpio_reg_wr(REG_GPIO_OE, 0);
}

void music_boot_elite(void)
{
    uint32_t *addr_p = (uint32_t *) ELITE_ENTRY_POINT;
    start_elite = (void (*)())ELITE_ENTRY_POINT;
    start_elite();
}

uint32_t boot_image_addr_computer()
{
    unsigned long image_physical_base = 0;
    ctrl_info_t *control_block_p, control_block;

    //check boot which image
    control_block_backup_check();
    control_block_p = (ctrl_info_t *) CONTROL_BLOCK_BASE;
    if (control_block_verification(control_block_p))
    {
        printf("control blcok crc bad, recover it\n\r");
        //at this time, backup1 and backup2 should be the same
        control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, (ctrl_info_t *) CONTROL_BLOCK_BACKUP1_ADDR);
    }

    if (control_block_p->upgrade_flag == UPGRADE_NSET)
    {
        if ((control_block_p->select_flag == IMAGE_A) && (control_block_p->imageA_valid_flag == IMAGE_VALID))
        {
            if (control_block_p->imageA_loader_count < MAX_LOAD_COUNT)
                image_physical_base = IMAGE_A_OFFSET;
        }
        else if ((control_block_p->select_flag == IMAGE_B) && (control_block_p->imageB_valid_flag == IMAGE_VALID))
        {
            if (control_block_p->imageB_loader_count < MAX_LOAD_COUNT)
                image_physical_base = IMAGE_B_OFFSET;
        }
        else
        {
            printf("No effective image exists!\n\r");
            return 0;
        }
    }
    else if (control_block_p->upgrade_flag == UPGRADE_SET)
    {
        memcpy((void *)&control_block, (void *)CONTROL_BLOCK_BASE, sizeof(control_block));
        control_block_p = &control_block;
        if ((control_block_p->select_flag == control_block_p->will_select_flag))
        {
            printf("select_flag equals to will_select_flag, bad!\n\r");
            return 0;
        }
        if (control_block_p->will_select_flag == IMAGE_A)
        {
            printf("loader check will_select_flag = IMAGE_A\n\r");
            control_block_p->imageA_loader_count++;
            if (control_block_p->imageA_loader_count >= MAX_LOAD_COUNT)
            {
                control_block_p->upgrade_flag = UPGRADE_NSET;
                control_block_p->will_select_flag = control_block_p->select_flag;
                control_block_p->image_is_decompressed = IMAGE_NOT_DECOMP;
                control_block_p->crc_32bit = control_block_crc_compute(&control_block);
                control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, (ctrl_info_t *) & control_block);
                image_physical_base = IMAGE_B_OFFSET;
            }
            control_block_p->crc_32bit = control_block_crc_compute(&control_block);
            control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, (ctrl_info_t *) & control_block);
            image_physical_base = IMAGE_A_OFFSET;
        }
        else
        {
            printf("loader check will_select_flag = IMAGE_B\n\r");
            control_block_p->imageB_loader_count++;
            if (control_block_p->imageB_loader_count >= MAX_LOAD_COUNT)
            {
                control_block_p->upgrade_flag = UPGRADE_NSET;
                control_block_p->will_select_flag = control_block_p->select_flag;
                control_block_p->image_is_decompressed = IMAGE_NOT_DECOMP;
                control_block_p->crc_32bit = control_block_crc_compute(&control_block);
                control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, (ctrl_info_t *) & control_block);
                image_physical_base = IMAGE_A_OFFSET;
            }
            control_block_p->crc_32bit = control_block_crc_compute(&control_block);
            control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, control_block_p);
            image_physical_base = IMAGE_B_OFFSET;
        }
    }
    else
    {
        printf("bug:program should not go here line:%d!\n\r", __LINE__);
        return 0;
    }
    printf("image_physical_base:%x\n\r", (uint32_t) image_physical_base);
    return image_physical_base;

}

static int zmodem_user_break(int delay_secs)
{
    int abort = 0, abort_key = 0;

    //printf("Press ctrl+r to remove configuration file\n\r");
    printf("Press ctrl+c to enter into command line\n\r");

    while ((delay_secs > 0) && (!abort))
    {
        int i;

        --delay_secs;

        /* delay 100 * 10ms */
        for (i = 0; !abort && i < 100; ++i)
        {
            if (serial_tstc())
            {
                abort_key = serial_getc();
                if (abort_key == ZM_USER_BREAK_INVALID_CONF_KEY)
                {
                    //spiflash_sector_erase(CONFIG_FILE_BASE-SPI_FLASH_BASE);
                    //printf("\n\rRemove configuration file completed!");
                    abort = 1;
                    delay_secs = 0;
                    goto OUT;
                }
                abort_key = 0;
            }
            udelay(10000);
        }

        printf("\b\b\b%2d ", delay_secs);
    }

  OUT:
    printf("\n\r");

    return abort_key;
}

void command_line_process(void)
{
    unsigned int read_size = 0;
    printf("\n%s", LOADER_PROMPT);
    while (1)
    {
        if (serial_tstc())
        {
            unsigned char c = serial_getc();
            if (0x8 == c)       // backspace
            {
#if 1
                if (serial_line_buf_len > 0)
                {
                    serial_putc(8);
                    serial_putc(32);
                    serial_putc(8);
                    serial_line_buf_len--;
                }
#endif
            }
            else if (0xd == c)
            {
                //serial_line_prompt();
                if (serial_line_buf_len <= MAX_SERIAL_BUF_LEN)
                    serial_buf[serial_line_buf_len] = '\0';
                else
                    serial_buf[MAX_SERIAL_BUF_LEN] = '\0';

                command_handler(serial_buf);

                printf("\n%s", LOADER_PROMPT);
                serial_line_buf_len = 0;
            }
            else if (c >= 32 && c <= 126)   // printable char
            {
                serial_putc(c);
                if (serial_line_buf_len < MAX_SERIAL_BUF_LEN)
                {
                    serial_buf[serial_line_buf_len] = c;
                }
                serial_line_buf_len++;
            }
            //serial_line_input_byte(c);
        }
        read_size = gmac_poll(packet_buf, VIVO_RX_BUF_SIZE);
        if (read_size)
        {
            agent_pkt_handle(packet_buf, read_size);
        }
    }
}

int main()
{
    unsigned long image_physical_base = 0;
    int abort_key = 0;
    unsigned int read_size = 0;
    ctrl_info_t *control_block_p, control_block;
    serial_init();
    printf("Loader version: %s\n\r", LOADER_VER_STRING);
    music_spi_init();
    gpio_config_init();
    board_get_partition_info();
    music_switch_main();
    gmac_init();

    if ((abort_key = zmodem_user_break(1)))
    {
    }

    if (abort_key)
        command_line_process();

    image_physical_base = boot_image_addr_computer();

    if (image_physical_base == IMAGE_A_OFFSET)
        printf("bootloader select image A!\n\r");
    else if (image_physical_base == IMAGE_B_OFFSET)
        printf("bootloader select image B!\n\r");
    else
    {
        printf("some serious problems occures, go into command line in loader\n\r");
        command_line_process();
    }

    if (whether_image_is_decompress() == 0)
    {
        if (uncompress_image(IMAGE_RUNNING_OFFSET, image_physical_base) == 0)
        {
            printf("decompressing image is successful\n\r");
        }
        else
        {
            printf("decompressing image is failed\n\r");
            command_line_process();
        }
    }

    watchdog_init();
    watchdog_start();
    printf("start boot elite at physical address:0x%x!\n\r", (int)image_physical_base);
    music_boot_elite();
    return 0;
}
