/* 
 * Copyright (c) 2012 Qualcomm Atheros, Inc.. 
 * All Rights Reserved. 
 * Qualcomm Atheros Confidential and Proprietary. 
 */

#include "music_spi.h"
#include "music_upgrade.h"
#include "mipsregs.h"

#define FAILED  -1

#define DEBUG 0
#if DEBUG
#define debug_upgrade_printf(...)    printf(__VA_ARGS__)
#else
#define debug_upgrade_printf(...)
#endif


#define call_sync()                                     \
	__asm__ __volatile__(                                   \
		".set noreorder\n\t"                            \
		"sync\n\t"                                     \
".set reorder")

#define tlb_read_indexed()                                     \
	__asm__ __volatile__(                                   \
		".set noreorder\n\t"                            \
		"tlbr\n\t"                                     \
".set reorder")

int printf(const char *fmt, ...);
int memcmp(const void *buf1, const void *buf2, unsigned int count);
int strcmp(const char *s1, const char *s2);
void *memcpy(void *dest, const void *src, int n);
uint32_t image_header_check(image_header_t * image_header)
{
#if 0
    board_info_t *board_info = (board_info_t *) BOARD_INFO_BASE;
//TODO
    if (strcmp(image_header->magic, (char *)"elite image") ||
        strcmp(image_header->oem_id, board_info->oem_id) || strcmp(image_header->oem_id, board_info->oem_id))
        return (FAILED);
#endif
    if (image_header->image_header_length != IMAGE_HEADER_SIZE)
    {
        printf("image_header_length:0x%x != IMAGE_HEADER_SIZE:0x%x\n\r",
               image_header->image_header_length, IMAGE_HEADER_SIZE);
        return (FAILED);
    }
    if (image_header->image_length > IMAGE_MAX_LENGTH)
    {
        printf("image_length:0x%x exceed IMAGE_MAX_LENGTH:0x%x\n\r", image_header->image_length, IMAGE_MAX_LENGTH);
        return (FAILED);
    }

    return 0;
}

#define call_sync()                                     \
	__asm__ __volatile__(                                   \
		".set noreorder\n\t"                            \
		"sync\n\t"                                     \
".set reorder")
uint32_t current_running_image()
{
#if 1
    ctrl_info_t *ctrl_info = (ctrl_info_t *) CONTROL_BLOCK_BASE;
    if (ctrl_info->upgrade_flag == UPGRADE_NSET)
        return ctrl_info->select_flag;
    else if (ctrl_info->upgrade_flag == UPGRADE_SET)
        return ctrl_info->will_select_flag;
#else
    uint32_t image_physical_address, start_addr;
    write_32bit_cp0_register(CP0_INDEX, 0);
    call_sync();
    tlb_read_indexed();
    call_sync();
    start_addr = (read_32bit_cp0_register(CP0_ENTRYLO0) & 0xffffffc0) << 6;
    image_physical_address = KSEG1ADDR(start_addr);
    debug_upgrade_printf("image_physical_address:0x%x\n\r", image_physical_address);

    if ((image_physical_address >= IMAGE_A_BASE) && (image_physical_address < IMAGE_A_BASE + IMAGE_MAX_LENGTH))
        return IMAGE_A;
    else if ((image_physical_address >= IMAGE_B_BASE) && (image_physical_address < IMAGE_B_BASE + IMAGE_MAX_LENGTH))
        return IMAGE_B;
    else
        printf("!!!!!!!!!we should not go here!!!!!!!\n\r");
    return FAILED;
#endif
}

uint32_t image_crc_check(uint32_t which_image)
{
    uint32_t len = 0;
    char *image_addr;
    char checksum = 0;
    image_header_t *image_header;
    uint32_t upgrade_image_base;

    if ((which_image != IMAGE_A) && (which_image != IMAGE_B))
    {
        printf("input which_image is not correct\n\r");
        return FAILED;
    }

    upgrade_image_base = (which_image == IMAGE_A) ? IMAGE_A_BASE : IMAGE_B_BASE;
    image_header = (image_header_t *) upgrade_image_base;

    if (image_header_check(image_header))
    {
        debug_upgrade_printf("%s head crc check failed!\n\r", (which_image == IMAGE_A) ? "IMAGE_A" : "IMAGE_B");
        return FAILED;
    }

    image_addr = (char *)(upgrade_image_base + IMAGE_HEADER_SIZE);
    debug_upgrade_printf("image_addr:0x%x, image_length:0x%x, image_checksum:0x%x\n\r",
                         (uint32_t) image_addr, image_header->image_length, image_header->image_checksum);
    len = 0;
    while (len < image_header->image_length)
    {
        checksum += *image_addr;
        image_addr++;
        len++;
    }
    if (image_header->image_checksum != checksum)
    {
        printf("computed checksum:0x%x != image_checksum:0x%x\n\r", checksum, image_header->image_checksum);
        return FAILED;
    }
    return 0;
}

uint32_t control_block_crc_compute(ctrl_info_t * control_info)
{
    uint32_t checksum, length, i;
    uint32_t *checksum_p = (uint32_t *) control_info;
    checksum = 0;
    length = CONTROL_BLOCK_LEN / 4 - 1;
    for (i = 0; i < length; i++)
    {
        checksum += *checksum_p;
        checksum_p++;
    }
    return checksum;
}

uint32_t control_block_verification(ctrl_info_t * control_info)
{
    uint32_t checksum, length, i;
    uint32_t *checksum_p = (uint32_t *) control_info;
    checksum = 0;
    length = CONTROL_BLOCK_LEN / 4 - 1;
    debug_upgrade_printf("control block@0x%x length:%d\n\r", (uint32_t) control_info, length);
    for (i = 0; i < length; i++)
    {
        checksum += *checksum_p;
        checksum_p++;
    }
    if (checksum != control_info->crc_32bit)
    {
        debug_upgrade_printf("control block@0x%x crc error!\n\r", (uint32_t) control_info);
        return FAILED;
    }
    else
        return 0;
}

uint32_t control_block_copy(ctrl_info_t * dest, ctrl_info_t * source)
{
    uint32_t dest_addr, src_addr;
    char temp_buf[MUSIC_SPI_PAGE_SIZE];
    dest_addr = (uint32_t) dest;
    src_addr = (uint32_t) source;
    if (!((dest_addr == CONTROL_BLOCK_BASE) || (dest_addr == CONTROL_BLOCK_BACKUP1_ADDR)
          || (dest_addr == CONTROL_BLOCK_BACKUP2_ADDR)))
    {
        printf("dest@0x%x is not correct\n\r", dest_addr);
        return FAILED;
    }
    if ((uint32_t) dest == (uint32_t) source)
    {
        printf("dest equals to source address!\n\r");
        return FAILED;
    }

    spiflash_sector_erase(dest_addr - SPI_FLASH_BASE);
    if ((src_addr == CONTROL_BLOCK_BASE) || (src_addr == CONTROL_BLOCK_BACKUP1_ADDR)
        || (src_addr == CONTROL_BLOCK_BACKUP2_ADDR))
    {
        debug_upgrade_printf("need to copy src_buf into memory\n\r");
        memcpy(temp_buf, (uint8_t *) (src_addr), MUSIC_SPI_PAGE_SIZE);
        spiflash_write_page(dest_addr - SPI_FLASH_BASE, (uint8_t *) (temp_buf), sizeof(*source));
    }
    else
        spiflash_write_page(dest_addr - SPI_FLASH_BASE, (uint8_t *) (src_addr), sizeof(*source));

    return 0;
}

uint32_t control_block_backup_check()
{
    ctrl_info_t *backup1, *backup2, *source, *dest;
    backup1 = (ctrl_info_t *) CONTROL_BLOCK_BACKUP1_ADDR;
    backup2 = (ctrl_info_t *) CONTROL_BLOCK_BACKUP2_ADDR;
    if (control_block_verification(backup1))
    {
        debug_upgrade_printf("backup1 crc error\n\r");
        source = backup2;
        dest = backup1;
        if (control_block_verification(backup2))
        {
            printf("both backup1 and backup2 crc error\n\r");
            return FAILED;
        }
    }
    else if (control_block_verification(backup2))
    {
        debug_upgrade_printf("backup2 crc error\n\r");
        source = backup1;
        dest = backup2;
    }
    else
    {
        //check whether backup1 and backup2 is the same. If not, use backup1 to restore backup2
        if (memcmp((void *)CONTROL_BLOCK_BACKUP1_ADDR, (void *)CONTROL_BLOCK_BACKUP2_ADDR, sizeof(ctrl_info_t)))
        {
            printf("backup1 and backup2 is not the same, but crc correct\n\r");
            source = backup1;
            dest = backup2;
        }
        else
            return 0;
    }

    if (control_block_copy(dest, source))
    {
        printf("control_block_copy failed!\n\r");
        return FAILED;
    }

    debug_upgrade_printf("recover control block from backup info\n\r");
    control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, source);
    return 0;
}

uint32_t control_block_backup_save()
{
    if (control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BACKUP1_ADDR, (ctrl_info_t *) CONTROL_BLOCK_BASE))
    {
        debug_upgrade_printf("copy CONTROL_BLOCK_BACKUP1_ADDR failed\n\r");
        return FAILED;
    }
    if (control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BACKUP2_ADDR, (ctrl_info_t *) CONTROL_BLOCK_BASE))
    {
        debug_upgrade_printf("copy CONTROL_BLOCK_BACKUP2_ADDR failed\n\r");
        return FAILED;
    }
    return 0;
}

uint32_t image_upgrade_identification_set()
{
    ctrl_info_t control_block;
    uint32_t image;
    uint32_t current_image = current_running_image();
    image = (current_image == IMAGE_A) ? IMAGE_B : IMAGE_A;

    memcpy((void *)&control_block, (void *)CONTROL_BLOCK_BASE, sizeof(control_block));
    control_block.upgrade_flag = UPGRADE_SET;
    control_block.will_select_flag = image;
    if (image == IMAGE_A)
    {
        control_block.imageA_loader_count = 0;
        control_block.imageA_valid_flag = IMAGE_UNVALID;
    }
    else
    {
        control_block.imageB_loader_count = 0;
        control_block.imageB_valid_flag = IMAGE_UNVALID;
    }
    control_block.image_is_decompressed = IMAGE_NOT_DECOMP;
    control_block.crc_32bit = control_block_crc_compute(&control_block);
    debug_upgrade_printf("control_block@0x%x crc_32bit:0x%x@0x%x\n\r", (uint32_t) & control_block,
                         control_block.crc_32bit, (uint32_t) & control_block.crc_32bit);
    return control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, (ctrl_info_t *) & control_block);
}

uint32_t image_upgrade_identification_check()
{
    ctrl_info_t control_block;
    memcpy((void *)&control_block, (void *)CONTROL_BLOCK_BASE, sizeof(control_block));
    if (control_block_verification(&control_block))
    {
        printf("control block crc error, this should not happen!\n\r");
        return FAILED;
    }
    if (control_block.upgrade_flag == UPGRADE_NSET)
        return 0;

    debug_upgrade_printf("upgrade_flag is set, will_select is %s\n\r",
                         (control_block.will_select_flag == IMAGE_A) ? "IMAGE_A" : "IMAGE_B");

    if (control_block.will_select_flag != current_running_image())
    {
        printf("current running image is not will_select_flag\n\r");
        return FAILED;
    }

    if (control_block.will_select_flag == IMAGE_A)
    {
        control_block.imageA_loader_count = 0;
        control_block.imageA_valid_flag = IMAGE_VALID;
    }
    else
    {
        control_block.imageB_loader_count = 0;
        control_block.imageB_valid_flag = IMAGE_VALID;
    }
    control_block.select_flag = control_block.will_select_flag;
    control_block.upgrade_flag = UPGRADE_NSET;
    control_block.image_is_decompressed = IMAGE_DECOMP;
    control_block.crc_32bit = control_block_crc_compute(&control_block);
    if (control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, (ctrl_info_t *) & control_block))
    {
        debug_upgrade_printf("copy CONTROL_BLOCK_BASE failed\n\r");
        return FAILED;
    }
    return control_block_backup_save();
}

uint32_t switchover_image_check()
{
    ctrl_info_t *control_block = (ctrl_info_t *) CONTROL_BLOCK_BASE;
    if (control_block_verification(control_block))
    {
        printf("CONTROL_BLOCK_BASE crc error, should not happen!\n\r");
        return FAILED;
    }
    if ((control_block->imageA_valid_flag != IMAGE_VALID) || (control_block->imageB_valid_flag != IMAGE_VALID))
    {
        debug_upgrade_printf("another image is not verified!\n\r");
        return FAILED;
    }

    if (image_crc_check(IMAGE_A))
    {
        printf("IMAGE_A check failed!\n\r");
        return FAILED;
    }

    if (image_crc_check(IMAGE_B))
    {
        printf("IMAGE_B check failed!\n\r");
        return FAILED;
    }

    return 0;
}

uint32_t switchover_image_operation()
{
    ctrl_info_t control_block;

    if (switchover_image_check())
        return FAILED;

    memcpy((void *)&control_block, (void *)CONTROL_BLOCK_BASE, sizeof(control_block));

    if (control_block.select_flag == IMAGE_A)
    {
        control_block.select_flag = IMAGE_B;
        debug_upgrade_printf("do switchover from IMAGE_A to IMAGE_B\n\r");
    }
    else
    {
        control_block.select_flag = IMAGE_A;
        debug_upgrade_printf("do switchover from IMAGE_B to IMAGE_A\n\r");
    }
    control_block.will_select_flag = control_block.select_flag;
    control_block.crc_32bit = control_block_crc_compute(&control_block);
    if (control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, (ctrl_info_t *) & control_block))
    {
        debug_upgrade_printf("copy CONTROL_BLOCK_BASE failed\n\r");
        return FAILED;
    }
    return control_block_backup_save();
}


void image_header_dump(image_header_t * image_header_p)
{
    if (image_header_check(image_header_p))
    {
        printf("image header crc error\n\r");
        return;
    }
    printf("magic: %s\n\r", image_header_p->magic);
    printf("image_version: %s\n\r", image_header_p->image_version);
    printf("build_time: %s\n\r", image_header_p->build_time);
    printf("oem_id: %s\n\r", image_header_p->oem_id);
    printf("product_id: %s\n\r", image_header_p->product_id);
    printf("image_header_length: 0x%x\n\r", image_header_p->image_header_length);
    printf("image_length: 0x%x\n\r", image_header_p->image_length);
    printf("image_checksum: 0x%x\n\r", image_header_p->image_checksum);
    printf("header_checksum: 0x%x\n\r", image_header_p->header_checksum);
}

void control_info_dump(ctrl_info_t * control_block_p)
{
    if (control_block_verification(control_block_p))
    {
        printf("control block crc error\n\r");
        return;
    }
    printf("upgrade_flag: %s\n\r", (control_block_p->upgrade_flag == UPGRADE_SET) ? "UPGRADE_SET" : "UPGRADE_NSET");
    printf("select_flag: %s\n\r", (control_block_p->select_flag == IMAGE_A) ? "IMAGE_A" : "IMAGE_B");
    printf("image has %s\n\r",
           (control_block_p->image_is_decompressed ==
            IMAGE_NOT_DECOMP) ? "not been decompressed" : "been decompressed");
    printf("will_select_flag: %s\n\r", (control_block_p->will_select_flag == IMAGE_A) ? "IMAGE_A" : "IMAGE_B");
    printf("imageA_loader_count: %d\n\r", control_block_p->imageA_loader_count);
    printf("imageA_valid_flag: %s\n\r",
           (control_block_p->imageA_valid_flag == IMAGE_VALID) ? "IMAGE_VALID" : "IMAGE_UNVALID");
    printf("imageB_loader_count: %d\n\r", control_block_p->imageB_loader_count);
    printf("imageB_valid_flag: %s\n\r",
           (control_block_p->imageB_valid_flag == IMAGE_VALID) ? "IMAGE_VALID" : "IMAGE_UNVALID");
}

void control_block_dump()
{
    ctrl_info_t *control_block_p = (ctrl_info_t *) CONTROL_BLOCK_BASE;
    printf("\ndump ctrl at address:0x%x\n\r", control_block_p);
    control_info_dump(control_block_p);
    control_block_p = (ctrl_info_t *) CONTROL_BLOCK_BACKUP1_ADDR;
    printf("\ndump ctrl at address:0x%x\n\r", control_block_p);
    control_info_dump(control_block_p);
    control_block_p = (ctrl_info_t *) CONTROL_BLOCK_BACKUP2_ADDR;
    printf("\ndump ctrl at address:0x%x\n\r", control_block_p);
    control_info_dump(control_block_p);

}

void ctrl_block_A_correct(void)
{
    ctrl_info_t control_block;

    memset((void *)&control_block, 0xff, sizeof(control_block));
    memcpy((void *)&control_block, (void *)CONTROL_BLOCK_BASE, sizeof(control_block));
    control_block.upgrade_flag = UPGRADE_NSET;
    control_block.select_flag = IMAGE_A;
    control_block.will_select_flag = IMAGE_A;
    control_block.imageA_loader_count = 0;
    control_block.imageA_valid_flag = IMAGE_VALID;
    control_block.imageB_loader_count = 0;
    control_block.imageB_valid_flag = IMAGE_UNVALID;
    control_block.image_is_decompressed = IMAGE_NOT_DECOMP;
    control_block.crc_32bit = control_block_crc_compute(&control_block);

    if (control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, (ctrl_info_t *) & control_block))
    {
        debug_upgrade_printf("copy CONTROL_BLOCK_BASE failed\n\r");
        return FAILED;
    }
    control_block_backup_save();
}

void ctrl_block_B_correct(void)
{
    ctrl_info_t control_block;

    memset((void *)&control_block, 0xff, sizeof(control_block));
    memcpy((void *)&control_block, (void *)CONTROL_BLOCK_BASE, sizeof(control_block));
    control_block.upgrade_flag = UPGRADE_NSET;
    control_block.select_flag = IMAGE_B;
    control_block.will_select_flag = IMAGE_B;
    control_block.imageA_loader_count = 0;
    control_block.imageA_valid_flag = IMAGE_UNVALID;
    control_block.imageB_loader_count = 0;
    control_block.imageB_valid_flag = IMAGE_VALID;
    control_block.image_is_decompressed = IMAGE_NOT_DECOMP;
    control_block.crc_32bit = control_block_crc_compute(&control_block);

    if (control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, (ctrl_info_t *) & control_block))
    {
        debug_upgrade_printf("copy CONTROL_BLOCK_BASE failed\n\r");
        return FAILED;
    }
    control_block_backup_save();
}

void ctrl_block_default(void)
{
    ctrl_info_t control_block;

    memset((void *)&control_block, 0xff, sizeof(control_block));
    memcpy((void *)&control_block, (void *)CONTROL_BLOCK_BASE, sizeof(control_block));
    control_block.upgrade_flag = UPGRADE_NSET;
    control_block.select_flag = IMAGE_A;
    control_block.will_select_flag = IMAGE_A;
    control_block.imageA_loader_count = 0;
    control_block.imageA_valid_flag = IMAGE_UNVALID;
    control_block.imageB_loader_count = 0;
    control_block.imageB_valid_flag = IMAGE_UNVALID;
    control_block.image_is_decompressed = IMAGE_NOT_DECOMP;
    control_block.crc_32bit = control_block_crc_compute(&control_block);

    if (control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, (ctrl_info_t *) & control_block))
    {
        debug_upgrade_printf("copy CONTROL_BLOCK_BASE failed\n\r");
        return FAILED;
    }
    control_block_backup_save();
    printf("ctrl block is not valid, using default setting\n\r");
}

#if 0
void image_is_decompress_mark(void)
{
    ctrl_info_t control_block;

    memset((void *)&control_block, 0xff, sizeof(control_block));
    memcpy((void *)&control_block, (void *)CONTROL_BLOCK_BASE, sizeof(control_block));
    control_block.image_is_decompressed = IMAGE_DECOMP;
    control_block.crc_32bit = control_block_crc_compute(&control_block);

    if (control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, (ctrl_info_t *) & control_block))
    {
        debug_upgrade_printf("copy CONTROL_BLOCK_BASE failed\n\r");
        return FAILED;
    }
    return;
}
#endif

uint32_t whether_image_is_decompress(void)
{
    ctrl_info_t control_block;
    ctrl_info_t *control_block_p;

    //control_block_backup_check();
    control_block_p = (ctrl_info_t *) CONTROL_BLOCK_BASE;
    if (control_block_verification(control_block_p))
    {
        printf("control blcok crc bad, recover it\n\r");
        //at this time, backup1 and backup2 should be the same
        control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, (ctrl_info_t *) CONTROL_BLOCK_BACKUP1_ADDR);
    }



    memcpy((void *)&control_block, (void *)CONTROL_BLOCK_BASE, sizeof(control_block));
    if (control_block.image_is_decompressed == IMAGE_DECOMP)
        return 1;
    else
        return 0;

}

uint32_t image_compress_offset(void)
{
    ctrl_info_t control_block;
    ctrl_info_t *control_block_p;
    uint32_t compress_offset = 0;

    //control_block_backup_check();
    control_block_p = (ctrl_info_t *) CONTROL_BLOCK_BASE;
    if (control_block_verification(control_block_p))
    {
        printf("control blcok crc bad, recover it\n\r");
        //at this time, backup1 and backup2 should be the same
        control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, (ctrl_info_t *) CONTROL_BLOCK_BACKUP1_ADDR);
    }



    memcpy((void *)&control_block, (void *)CONTROL_BLOCK_BASE, sizeof(control_block));

    if (control_block.will_select_flag == IMAGE_A)
        compress_offset = IMAGE_A_OFFSET;
    else
        compress_offset = IMAGE_B_OFFSET;
    printf("image_compress_offset:0x%x\n", compress_offset);
    return compress_offset;

}

uint32_t whether_can_upgrade_image()
{
    ctrl_info_t *control_block_p;

    control_block_p = (ctrl_info_t *) CONTROL_BLOCK_BASE;
    if (control_block_p->select_flag == current_running_image())
        return 1;
    else
        return 0;
}

void image_upgrade(void)
{
    if (whether_can_upgrade_image())
        image_upgrade_identification_set();
    else
        printf("image_upgrade is forbiden!\n\r");
}

void image_upgrade_done(void)
{
    image_upgrade_identification_check();
}

int invalidate_image(uint32_t image)
{
    ctrl_info_t control_block;

    memcpy((void *)&control_block, (void *)CONTROL_BLOCK_BASE, sizeof(control_block));

    if (control_block_verification(&control_block))
    {
        ctrl_block_default();
    }
    if (image == IMAGE_A)
    {
        control_block.imageA_loader_count = 0;
        control_block.imageA_valid_flag = IMAGE_UNVALID;
    }
    else
    {
        control_block.imageB_loader_count = 0;
        control_block.imageB_valid_flag = IMAGE_UNVALID;
    }
    control_block.crc_32bit = control_block_crc_compute(&control_block);
    control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, (ctrl_info_t *) & control_block);
    control_block_backup_save();
    return 0;
}

int control_block_update(uint32_t image)
{
    ctrl_info_t control_block;

    if (image_crc_check(image))
    {
        printf("image crc bad!\n\r");
        return -1;
    }

    memcpy((void *)&control_block, (void *)CONTROL_BLOCK_BASE, sizeof(control_block));
    control_block.upgrade_flag = UPGRADE_SET;
    control_block.will_select_flag = image;
    control_block.select_flag = (image == IMAGE_A) ? IMAGE_B : IMAGE_A;

    if (image == IMAGE_A)
    {
        control_block.imageA_loader_count = 0;
        control_block.imageA_valid_flag = IMAGE_UNVALID;
    }
    else
    {
        control_block.imageB_loader_count = 0;
        control_block.imageB_valid_flag = IMAGE_UNVALID;
    }
    control_block.image_is_decompressed = IMAGE_NOT_DECOMP;
    control_block.crc_32bit = control_block_crc_compute(&control_block);
    control_block_copy((ctrl_info_t *) CONTROL_BLOCK_BASE, (ctrl_info_t *) & control_block);
    control_block_backup_save();
    return 0;
}
