/*
 * Copyright (c) 2012 Qualcomm Atheros, Inc.
 * All Rights Reserved.
 * Qualcomm Atheros Confidential and Proprietary.
 * Notifications and licenses are retained for attribution purposes only
 */

/*
 * Copyright (c) 2004, Adam Dunkels.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * This file is part of the Contiki operating system.
 *
 * Author: Adam Dunkels <adam@sics.se>
 *
 * $Id: //depot/sw/switch/app/elite/trunk/elite/source/webserver/httpd.c#24 $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "contiki-net.h"
#include "sys_adapt.h"
#include "webserver-nogui.h"
#include "httpd-fs.h"
#include "lib/petsciiconv.h"
#include "http-strings.h"

#include "httpd.h"
#include "elite_init.h"
#include "elite_sys.h"
#include "http_upgrade.h"
#include "http_restore_cfg.h"
#include "asp_submit.h"
#include "asp_fun_fdb.h"
#include "web_config_dispatch.h"

#ifndef WEBSERVER_CONF_CGI_CONNS
#define CONNS UIP_CONNS
#else
#define CONNS WEBSERVER_CONF_CGI_CONNS
#endif

#define STATE_WAITING 0
#define STATE_OUTPUT  1

#define SEND_STRING(s, str) PSOCK_SEND(s, (uint8_t *)str, (unsigned int)strlen(str))
MEMB(conns, struct httpd_state, CONNS);

#define ISO_nl      0x0a
#define ISO_space   0x20
#define ISO_bang    0x21
#define ISO_percent 0x25
#define ISO_period  0x2e
#define ISO_slash   0x2f
#define ISO_colon   0x3a

#define CFG_FILE_HEADER_LEN 59

extern UINT8 compose_config(UINT8 *, UINT16 *);

UINT8 cfgPacket[4096] = {0};
struct cfg_info config;

/*---------------------------------------------------------------------------*/
static UINT16 is_config_file(char *filename)
{
    char file[HTTPD_FILE_NAME_LEN + 1] = {0};
    char *ptr;
	
    strcpy(file, filename);
    ptr = strrchr(file, ISO_period);
    if (ptr != NULL)
    {
        if (strncmp(ptr, http_cfg, 4) == 0)
        {
            return 1;
        }
    }

    return 0;
}

/*---------------------------------------------------------------------------*/
static void assemble_config_file_packet(struct httpd_cfg_file *file)
{
    UINT16 cfgPktLen = 0;
    UINT8 *cfgPacket_p =cfgPacket; 

    memcpy(cfgPacket, http_config_header, CFG_FILE_HEADER_LEN);
    cfgPacket_p += CFG_FILE_HEADER_LEN;
    compose_config(cfgPacket_p, &cfgPktLen);
    file->data = cfgPacket;
    file->len = cfgPktLen + CFG_FILE_HEADER_LEN;
}

/*---------------------------------------------------------------------------*/
static unsigned short cfg_generate(void *state)
{
    struct cfg_info *cfg = (struct cfg_info *)state;

    if (cfg->file.len > uip_mss())
    {
        cfg->len = uip_mss();
    }
    else
    {
        cfg->len = cfg->file.len;
    }
    memcpy(uip_appdata, cfg->file.data, cfg->len);

    return cfg->len;
}

/*---------------------------------------------------------------------------*/
static unsigned short generate(void *state)
{
    struct httpd_state *s = (struct httpd_state *)state;

    if (s->file.len > uip_mss())
    {
        s->len = uip_mss();
    }
    else
    {
        s->len = s->file.len;
    }
    memcpy(uip_appdata, s->file.data, s->len);

    return s->len;
}

/*---------------------------------------------------------------------------*/
static PT_THREAD(send_config_file(struct httpd_state *s, struct cfg_info *cfg))
{
    PSOCK_BEGIN(&s->sout);

	//printf("file length = [%d]\r\n", cfg->file.len);
    do
    {
        PSOCK_GENERATOR_SEND(&s->sout, cfg_generate, cfg);
        cfg->file.len -= cfg->len;
        cfg->file.data += cfg->len;
    }
    while (cfg->file.len > 0);

    memset(cfgPacket, 0, 4096);
    PSOCK_END(&s->sout);
}

/*---------------------------------------------------------------------------*/
static PT_THREAD(send_file(struct httpd_state *s))
{
    PSOCK_BEGIN(&s->sout);

    do
    {
        PSOCK_GENERATOR_SEND(&s->sout, generate, s);
        s->file.len -= s->len;
        s->file.data += s->len;
    }
    while (s->file.len > 0);

    PSOCK_END(&s->sout);
}

/*---------------------------------------------------------------------------
 * This function convert source html file script line to real,
 * line by line, and copy to uip_appdata, until it is full.
 * The source html file length maybe not equal to real html file.
 * The return value copiedLen is length of copied bytes to dest buffer.
 */
static unsigned short html_generate(void *state)
{
    struct httpd_state *s = (struct httpd_state *)state;
    int copiedLen;

    s->len = 0;
    copiedLen = asp_file_copy((char *)uip_appdata, uip_mss(), s->file.data, s->file.len, &(s->len), &(s->asp));

    return copiedLen;
}

/*---------------------------------------------------------------------------*/
static PT_THREAD(send_html_file(struct httpd_state *s))
{
    PSOCK_BEGIN(&s->sout);

    /* init some ASP member */
    s->asp.curLine = 0;
    s->asp.lineMode = ASP_LINEMODE_NORMAL;
//  s->asp.fileName = s->filename;
    s->asp.curLineMark = 0;
    asp_cur_line_key_reset(&s->asp);

    do
    {
        PSOCK_GENERATOR_SEND(&s->sout, html_generate, s);
        s->file.len -= s->len;
        s->file.data += s->len;
    }
    while (s->file.len > 0);

    PSOCK_END(&s->sout);
}

static PT_THREAD(send_headers(struct httpd_state *s, const char *statushdr))
{
/* gcc warning if not initialized.
 * If initialized, minimal-net platform segmentation fault if not static...
 */
    static const char *ptr = NULL;

    PSOCK_BEGIN(&s->sout);

    SEND_STRING(&s->sout, statushdr);

    ptr = strrchr(s->filename, ISO_period);
    if (ptr == NULL)
    {
        ptr = http_content_type_binary;
    }
    else if (strncmp(http_html, ptr, 5) == 0 ||
             strncmp(http_shtml, ptr, 6) == 0 || strncmp(http_htm, ptr, 4) == 0 || strncmp(http_js, ptr, 3) == 0)
    {
        ptr = http_content_type_html;
    }
    else if (strncmp(http_css, ptr, 4) == 0)
    {
        ptr = http_content_type_css;
    }
    else if (strncmp(http_png, ptr, 4) == 0)
    {
        ptr = http_content_type_png;
    }
    else if (strncmp(http_gif, ptr, 4) == 0)
    {
        ptr = http_content_type_gif;
    }
    else if (strncmp(http_jpg, ptr, 4) == 0)
    {
        ptr = http_content_type_jpg;
    }
    else
    {
        ptr = http_content_type_plain;
    }
    SEND_STRING(&s->sout, ptr);
    PSOCK_END(&s->sout);
}

static unsigned short _send_redirect_to_login(void *state)
{
    char *dstBuf = (char *)uip_appdata;
    int stringLen;
    int dstCopiedLen = 0;
    char *loginHtm = "<html><body><script language=JavaScript>\n"
        "parent.location.href='/login.htm';\n" "</script></body></html>\n";

    stringLen = strlen(http_header_200);
    memcpy(dstBuf, http_header_200, stringLen);
    dstCopiedLen += stringLen;

    stringLen = strlen(http_content_type_html);
    memcpy((dstBuf + dstCopiedLen), http_content_type_html, stringLen);
    dstCopiedLen += stringLen;

    stringLen = strlen(loginHtm);
    memcpy((dstBuf + dstCopiedLen), loginHtm, stringLen);
    dstCopiedLen += stringLen;

    return dstCopiedLen;
}

static unsigned short _send_upgrade_process(void *state)
{
    char *dstBuf = (char *)uip_appdata;
    int stringLen;
    int dstCopiedLen = 0;
    int percent = 0;
    char processText[10] = { 0 };

    http_upgrade_percent_get(&percent);

    sprintf(processText, "%d", percent);

    stringLen = strlen(http_header_200);
    memcpy(dstBuf, http_header_200, stringLen);
    dstCopiedLen += stringLen;

    stringLen = strlen(http_content_type_plain);
    memcpy((dstBuf + dstCopiedLen), http_content_type_plain, stringLen);
    dstCopiedLen += stringLen;

    stringLen = strlen(processText);
    memcpy((dstBuf + dstCopiedLen), processText, stringLen);
    dstCopiedLen += stringLen;

    return dstCopiedLen;
}

static PT_THREAD(send_redirect_to_login(struct httpd_state *s))
{
    PSOCK_BEGIN(&s->sout);

    PSOCK_GENERATOR_SEND(&s->sout, _send_redirect_to_login, s);

    PSOCK_END(&s->sout);
}

static PT_THREAD(send_upgrade_process(struct httpd_state *s))
{
    PSOCK_BEGIN(&s->sout);

    PSOCK_GENERATOR_SEND(&s->sout, _send_upgrade_process, s);

    PSOCK_END(&s->sout);
}

static unsigned short _send_submit_reply(void *state)
{
    struct httpd_state *s = (struct httpd_state *)state;
    char *dstBuf = (char *)uip_appdata;
    int stringLen;
    int dstCopiedLen = 0;

    stringLen = strlen(http_header_200);
    memcpy(dstBuf, http_header_200, stringLen);
    dstCopiedLen += stringLen;

    if (s->setCookie)
    {
        char *cookie;
        sys_newCookie(&cookie);
        WEB_DEBUG("Set-Cookie: SID=%s\r\n", cookie);
        sprintf((dstBuf + dstCopiedLen), "Set-Cookie: SID=%s\r\n", cookie);
        dstCopiedLen += 18 + ELITE_SYS_COOKIE_MAX_LEN;
    }

    stringLen = strlen(http_content_type_html);
    memcpy((dstBuf + dstCopiedLen), http_content_type_html, stringLen);
    dstCopiedLen += stringLen;

    dstCopiedLen += asp_reload_page((dstBuf + dstCopiedLen), (uip_mss() - dstCopiedLen), &(s->asp));

    return dstCopiedLen;
}

static PT_THREAD(send_submit_reply(struct httpd_state *s))
{
    PSOCK_BEGIN(&s->sout);

    PSOCK_GENERATOR_SEND(&s->sout, _send_submit_reply, s);

    PSOCK_END(&s->sout);
}

static PT_THREAD(handle_output(struct httpd_state *s))
{
    PT_BEGIN(&s->outputpt);

	if (strcmp(s->asp.reloadUrl, "maintenance_firmware_upgrade.htm") == 0)
	{
		PT_WAIT_THREAD(&s->outputpt, send_submit_reply(s));
	}
    else if ((!s->authed) && (strstr(s->filename, ".htm") != NULL) && (strcmp(s->filename, "/login.htm") != 0))
    {
        PT_WAIT_THREAD(&s->outputpt, send_redirect_to_login(s));
    }
    else if (s->asp.method == HTTP_METHOD_GET)
    {
        if (strcmp(s->filename, "/system_upprocess") == 0)
        {
            PT_WAIT_THREAD(&s->outputpt, send_upgrade_process(s));
        }
        else if (strcmp(s->filename, "/upgrade_abort") == 0)
        {
            webserver_log("== http upgrade abort cause of user called /upgrade_abort ==\n");
            http_upgrade_infor_clear();
            PT_WAIT_THREAD(&s->outputpt, send_headers(s, http_header_200));
        }
	else if (is_config_file(s->filename))
        {
	        memset(&config, 0, sizeof(struct cfg_info));
            assemble_config_file_packet(&config.file);         
	        PT_WAIT_THREAD(&s->outputpt, send_config_file(s, &config));
        }
        else if (!httpd_fs_open(s->filename, &s->file))
        {
            strcpy(s->filename, http_404_html);
            httpd_fs_open(s->filename, &s->file);
            PT_WAIT_THREAD(&s->outputpt, send_headers(s, http_header_404));
            PT_WAIT_THREAD(&s->outputpt, send_file(s));
        }
        else
        {
            PT_WAIT_THREAD(&s->outputpt, send_headers(s, http_header_200));

            {
                char *ptr;
                ptr = strrchr(s->filename, ISO_period);
                if (ptr != NULL)
                {
                    if ((strncmp(ptr, http_html, 5) == 0) ||
                        (strncmp(ptr, http_htm, 4) == 0) || (strncmp(ptr, http_js, 3) == 0))
                    {
                        PT_INIT(&s->scriptpt);
                        PT_WAIT_THREAD(&s->outputpt, send_html_file(s));
                    }
                    else
                    {
                        PT_WAIT_THREAD(&s->outputpt, send_file(s));
                    }
                }
            }
        }
    }
    else if (s->asp.method == HTTP_METHOD_POST)
    {
        PT_WAIT_THREAD(&s->outputpt, send_submit_reply(s));
    }
    else
    {
        /* do nothing */
    }

    PSOCK_CLOSE(&s->sout);
    PT_END(&s->outputpt);
}

static PT_THREAD(analysis_url(struct httpd_state *s))
{
    PSOCK_BEGIN(&s->sin);

    /* analysis the method */
    s->asp.method = HTTP_METHOD_UNKNOWN;
    PSOCK_READTO(&s->sin, ISO_space);
    if (strncmp(s->inputbuf, http_get, 4) == 0)
    {
        s->asp.method = HTTP_METHOD_GET;
    }
    else if (strncmp(s->inputbuf, "POST ", 5) == 0)
    {
        s->asp.method = HTTP_METHOD_POST;
    }
    else
    {
        PSOCK_CLOSE_EXIT(&s->sin);
    }

    /* analysis the URL */
    PSOCK_READTO(&s->sin, ISO_space);
    if (s->inputbuf[0] != ISO_slash)
    {
        PSOCK_CLOSE_EXIT(&s->sin);
    }

    if (s->inputbuf[1] == ISO_space)
    {
        strncpy(s->filename, http_index_htm, sizeof(s->filename));
    }
    else
    {
        char *searchPointer = NULL;

        s->inputbuf[PSOCK_DATALEN(&s->sin) - 1] = 0;
        searchPointer = strchr(s->inputbuf, (int)'?');
        if (searchPointer == NULL)
        {
            strncpy(s->filename, &s->inputbuf[0], sizeof(s->filename));
            asp_query_clean(&(s->asp.env));
        }
        else
        {
            strncpy(s->filename, &s->inputbuf[0], MIN(searchPointer - s->inputbuf, sizeof(s->filename)));

            /* parse query: xxx.htm?a1=b1&a2=b2 */
            if (asp_query_parse(&(s->asp.env), searchPointer + 1))
            {
                ASP_ERROR("asp_query_parse() return error\r\n");
            }
        }
    }

    /* log and print */
    petsciiconv_topetscii(s->filename, sizeof(s->filename));
    {
        /* action is string like: GET/POST file_name?a1=b1&a2=b2 */
        char action[5 + HTTPD_FILE_NAME_LEN + 1 + sizeof(ASP_ENV)] = { 0 };

        if (s->asp.method == HTTP_METHOD_GET)
        {
            strcat(action, "GET ");
        }
        else if (s->asp.method == HTTP_METHOD_POST)
        {
            strcat(action, "POST ");
        }

        strcat(action, s->filename);

        asp_print_env(&(s->asp.env), action + strlen(action));

        webserver_log_file(&uip_conn->ripaddr, action);
    }
    petsciiconv_toascii(s->filename, sizeof(s->filename));

    PSOCK_END(&s->sin);
}

static PT_THREAD(analysis_rest_header(struct httpd_state *s))
{
    PSOCK_BEGIN(&s->sin);

    while (1)
    {
        PSOCK_READTO(&s->sin, ISO_nl);

        if (strncmp(s->inputbuf, "\r\n", 2) == 0)
        {
            /* this is last line of header */
            break;
        }

        {
            char *searchPointer = NULL;

            searchPointer = strpbrk(s->inputbuf, "\r\n");
            if (searchPointer != NULL)
            {
                *searchPointer = 0;
            }
        }

        if (strncmp(s->inputbuf, "Content-Length:", 15) == 0)
        {
            char *digitBegin = s->inputbuf + 15;

            while (*digitBegin && (*digitBegin == ' ')) /* skip leading spaces */
                digitBegin++;

            s->asp.postcontentlen = atoi(digitBegin);
        }
        else if (strncmp(s->inputbuf, "Cookie: SID=", 12) == 0)
        {
            char *cookieBegin = s->inputbuf + 12;

            s->authed = sys_checkCookie(cookieBegin);
        }
        else
        {
            /* do nothing */
        }
    }

    PSOCK_END(&s->sin);
}

static PT_THREAD(handle_input(struct httpd_state *s))
{
    PT_BEGIN(&s->inputpt);

    /* analysis url */
    PT_WAIT_THREAD(&s->inputpt, analysis_url(s));

    /* handle upgrade post request */
	//printf("s->asp.method:%d,s->filename:%s\n",s->asp.method,s->filename);
    if ((s->asp.method == HTTP_METHOD_POST) && (strcmp(s->filename, "/config/maintenance_firmware_upgrade.htm") == 0))		
    {
        //PT_INIT(&s->upgrade.upgradePt);
        if (http_upgrade_is_running())
        {
            HTTP_DEBUG("another session's doing updating, so upgrading and switchover is forbidden\n");
        }
        else
        {
            if (asp_get_env(&s->asp.env, "method"))
            {
                goto Asp;
            }
            else
            {
                PT_WAIT_THREAD(&s->inputpt, http_upgrade(s));
            }
        }

    }
	
    else if ((s->asp.method == HTTP_METHOD_POST) && (strcmp(s->filename, "/config/maintenance_restore_configuration.htm") == 0))		
    {
        //PT_INIT(&s->upgrade.upgradePt);
        if (http_restore_cfg_is_running())
        {
            HTTP_DEBUG("another session's doing updating, so uploading configuration is forbidden\n");
        }
        else
        {
            if (asp_get_env(&s->asp.env, "method"))
            {
                goto Asp;
            }
            else
            {
                PT_WAIT_THREAD(&s->inputpt, http_cfgrestore(s));
            }
        }
    }
	
    else
    {
      Asp:
        ;
        /* read and analysis rest of header */
        PT_WAIT_THREAD(&s->inputpt, analysis_rest_header(s));

#if ELITE_MODULE_WEB_NO_LOGIN
        s->authed = TRUE;
#endif

        /* handle post content */
        if ((s->asp.method == HTTP_METHOD_POST) && ((s->authed) || (strcmp(s->filename, "/login.htm") == 0)))
        {
            PT_WAIT_THREAD(&s->inputpt, asp_submit(s));
        }
    }

    s->state = STATE_OUTPUT;
    PT_END(&s->inputpt);
}

static void handle_connection(struct httpd_state *s)
{
    if (s->state == STATE_WAITING)
    {
        handle_input(s);
    }

    if (s->state == STATE_OUTPUT)
    {
        handle_output(s);
    }
}

static void asp_init(struct httpd_state *s)
{
    memset(&(s->asp), 0, sizeof(ASP_INFO_T));
    s->asp.pHttpdState = (void *)s;
    s->asp.fileName = s->filename;
    PT_INIT(&s->asp.submitPt);
    PT_INIT(&s->asp.submitFunPt);

    return;
}

void httpd_appcall(void *state)
{
    struct httpd_state *s = (struct httpd_state *)state;

    if (uip_closed() || uip_aborted() || uip_timedout())
    {
        if (s != NULL)
        {
            memb_free(&conns, s);
        }
    }
    else if (uip_connected())
    {
        s = (struct httpd_state *)memb_alloc(&conns);
        if (s == NULL)
        {
            uip_abort();
            return;
        }
        memset(s, 0, sizeof(struct httpd_state));
        tcp_markconn(uip_conn, s);
        PSOCK_INIT(&s->sin, (uint8_t *) s->inputbuf, sizeof(s->inputbuf) - 1);
        PSOCK_INIT(&s->sout, (uint8_t *) s->inputbuf, sizeof(s->inputbuf) - 1);
        PT_INIT(&s->inputpt);
        PT_INIT(&s->outputpt);
        s->state = STATE_WAITING;
        /*    timer_set(&s->timer, CLOCK_SECOND * 100); */
        s->timer = 0;

        asp_init(s);

        handle_connection(s);
    }
    else if (s != NULL)
    {
        if (uip_poll())
        {
            ++s->timer;
            if (s->timer >= 20)
            {
                if (s->upgrading)
                {
                    http_upgrade_infor_clear();
                    webserver_log("== http upgrade abort cause of uip_poll() timed out ==\n");
                }

                uip_abort();
                memb_free(&conns, s);
            }
        }
        else
        {
            s->timer = 0;
        }
        handle_connection(s);
    }
    else
    {
        uip_abort();
    }
}

void httpd_init(void)
{
    tcp_listen(UIP_HTONS(80));
    memb_init(&conns);
}

#if UIP6_CONF_IPV6
/*---------------------------------------------------------------------------*/
uint8_t httpd_sprint_ip6(uip_ip6addr_t addr, char *result)
{
    unsigned char i = 0;
    unsigned char zerocnt = 0;
    unsigned char numprinted = 0;
    char *starting = result;

    *result++ = '[';
    while (numprinted < 8)
    {
        if ((addr.u16[i] == 0) && (zerocnt == 0))
        {
            while (addr.u16[zerocnt + i] == 0)
                zerocnt++;
            if (zerocnt == 1)
            {
                *result++ = '0';
                //numprinted++;
                //break;
            }
            i += zerocnt;
            numprinted += zerocnt;
        }
        else
        {
            result += sprintf(result, "%x", (unsigned int)(uip_ntohs(addr.u16[i])));
            i++;
            numprinted++;
        }
        if (numprinted != 8)
            *result++ = ':';
    }
    *result++ = ']';
    *result = 0;
    return (result - starting);
}
#endif                          /* UIP_CONF_IPV6 */
/*---------------------------------------------------------------------------*/
