/*
 * Copyright (c) 2012 Qualcomm Atheros, Inc.
 * All Rights Reserved.
 * Qualcomm Atheros Confidential and Proprietary.
 * Notifications and licenses are retained for attribution purposes only
 */

/*
 * Copyright (c) 2001-2005, Adam Dunkels.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This file is part of the uIP TCP/IP stack.
 *
 * $Id: //depot/sw/switch/app/elite/trunk/elite/source/webserver/httpd.h#13 $
 *
 */

#ifndef __HTTPD_H__
#define __HTTPD_H__

#ifdef __cplusplus
extern "C" {
#endif    /* __cplusplus */



#include "contiki-net.h"
#include "httpd-fs.h"
#include "asp.h"

#define __HTTP_DEBUG

#ifdef __HTTP_DEBUG
#define HTTP_DEBUG(fmt, ...)                                        \
    do                                                              \
    {                                                               \
        _elite_debug(ELITE_DEBUG_LEVEL_TRACE, MODULE_WEB,           \
                     __FUNCTION__, __LINE__, fmt, ##__VA_ARGS__);   \
    } while (0)
#else
#define HTTP_DEBUG(fmt, arg...) do{}while(0)
#endif

#define HTTP_ERROR(fmt, ...)                                         \
    do                                                              \
    {                                                               \
        _elite_debug(ELITE_DEBUG_LEVEL_ERROR, MODULE_WEB,           \
                     __FUNCTION__, __LINE__, fmt, ##__VA_ARGS__);   \
    } while (0)

#define HTTP_LOG(fmt, ...)                                           \
    do                                                              \
    {                                                               \
        _elite_debug(ELITE_DEBUG_LEVEL_LOG, MODULE_WEB,             \
                     __FUNCTION__, __LINE__, fmt, ##__VA_ARGS__);   \
    } while (0)

#define HTTPD_INPUT_BUF_LEN  256
#define HTTPD_FILE_NAME_LEN  63

struct httpd_state
{
    unsigned char timer;
    struct psock sin, sout;
    struct pt inputpt, outputpt, scriptpt;
    char inputbuf[HTTPD_INPUT_BUF_LEN + 1];
    char filename[HTTPD_FILE_NAME_LEN + 1];
    char state;
    struct httpd_fs_file file;
    int len;
    char *scriptptr;
    int scriptlen;
    union
    {
        unsigned short count;
        void *ptr;
    } u;

    BOOL authed;
    BOOL setCookie;
    ASP_INFO_T asp;             /* asp information */
    BOOL upgrading;
};

struct cfg_info
{
    struct httpd_cfg_file file;
    int len;
};

void httpd_init(void);
void httpd_appcall(void *state);

#if UIP6_CONF_IPV6
uint8_t httpd_sprint_ip6(uip_ip6addr_t addr, char *result);
#endif                          /* UIP6_CONF_IPV6 */


#ifdef __cplusplus
}
#endif    /* __cplusplus */

#endif                          /* __HTTPD_H__ */
